###########################
# Visual modality utilities
###########################

from builtins import range
from collections import defaultdict

from fonduer.candidates.mentions import Ngrams
from fonduer.utils.data_model_utils.utils import _to_span, _to_spans
from fonduer.utils.utils import tokens_to_ngrams
from fonduer.utils.utils_visual import (
    bbox_from_sentence,
    bbox_from_span,
    bbox_horz_aligned,
    bbox_vert_aligned,
    bbox_vert_aligned_center,
    bbox_vert_aligned_left,
    bbox_vert_aligned_right,
)


def get_page(mention):
    """Return the page number of the given mention.

    If a candidate is passed in, this returns the page of its first Mention.

    :param mention: The Mention to get the page number of.
    :rtype: integer
    """
    span = _to_span(mention)
    return span.get_attrib_tokens("page")[0]


def is_horz_aligned(c):
    """Return True if all the components of c are horizontally aligned.

    Horizontal alignment means that the bounding boxes of each Mention of c
    shares a similar y-axis value in the visual rendering of the document.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_horz_aligned(
                bbox_from_span(_to_span(c[i])), bbox_from_span(_to_span(c[0]))
            )
            for i in range(len(c))
        ]
    )


def is_vert_aligned(c):
    """Return true if all the components of c are vertically aligned.

    Vertical alignment means that the bounding boxes of each Mention of c
    shares a similar x-axis value in the visual rendering of the document.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_vert_aligned(
                bbox_from_span(_to_span(c[i])), bbox_from_span(_to_span(c[0]))
            )
            for i in range(len(c))
        ]
    )


def is_vert_aligned_left(c):
    """Return true if all components are vertically aligned on their left border.

    Vertical alignment means that the bounding boxes of each Mention of c
    shares a similar x-axis value in the visual rendering of the document. In
    this function the similarity of the x-axis value is based on the left
    border of their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_vert_aligned_left(
                bbox_from_span(_to_span(c[i])), bbox_from_span(_to_span(c[0]))
            )
            for i in range(len(c))
        ]
    )


def is_vert_aligned_right(c):
    """Return true if all components vertically aligned on their right border.

    Vertical alignment means that the bounding boxes of each Mention of c
    shares a similar x-axis value in the visual rendering of the document. In
    this function the similarity of the x-axis value is based on the right
    border of their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_vert_aligned_right(
                bbox_from_span(_to_span(c[i])), bbox_from_span(_to_span(c[0]))
            )
            for i in range(len(c))
        ]
    )


def is_vert_aligned_center(c):
    """Return true if all the components are vertically aligned on their center.

    Vertical alignment means that the bounding boxes of each Mention of c
    shares a similar x-axis value in the visual rendering of the document. In
    this function the similarity of the x-axis value is based on the center of
    their bounding boxes.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_vert_aligned_center(
                bbox_from_span(_to_span(c[i])), bbox_from_span(_to_span(c[0]))
            )
            for i in range(len(c))
        ]
    )


def same_page(c):
    """Return true if all the components of c are on the same page of the document.

    Page numbers are based on the PDF rendering of the document. If a PDF file is
    provided, it is used. Otherwise, if only a HTML/XML document is provided, a
    PDF is created and then used to determine the page number of a Mention.

    :param c: The candidate to evaluate
    :rtype: boolean
    """
    return all(
        [
            _to_span(c[i]).sentence.is_visual()
            and bbox_from_span(_to_span(c[i])).page
            == bbox_from_span(_to_span(c[0])).page
            for i in range(len(c))
        ]
    )


def get_horz_ngrams(
    mention, attrib="words", n_min=1, n_max=1, lower=True, from_sentence=True
):
    """Return all ngrams which are visually horizontally aligned with the Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention to evaluate
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :param from_sentence: If True, returns ngrams from any horizontally aligned
        Sentences, rather than just horizontally aligned ngrams themselves.
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in _get_direction_ngrams(
            "horz", span, attrib, n_min, n_max, lower, from_sentence
        ):
            yield ngram


def get_vert_ngrams(
    mention, attrib="words", n_min=1, n_max=1, lower=True, from_sentence=True
):
    """Return all ngrams which are visually vertivally aligned with the Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention to evaluate
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :param from_sentence: If True, returns ngrams from any horizontally aligned
        Sentences, rather than just horizontally aligned ngrams themselves.
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in _get_direction_ngrams(
            "vert", span, attrib, n_min, n_max, lower, from_sentence
        ):
            yield ngram


def _get_direction_ngrams(direction, c, attrib, n_min, n_max, lower, from_sentence):
    # TODO: this currently looks only in current table;
    #   precompute over the whole document/page instead
    bbox_direction_aligned = (
        bbox_vert_aligned if direction == "vert" else bbox_horz_aligned
    )
    ngrams_space = Ngrams(n_max=n_max, split_tokens=[])
    f = (lambda w: w.lower()) if lower else (lambda w: w)
    spans = _to_spans(c)
    for span in spans:
        if not span.sentence.is_tabular() or not span.sentence.is_visual():
            continue
        for sentence in span.sentence.table.sentences:
            if from_sentence:
                if (
                    bbox_direction_aligned(
                        bbox_from_sentence(sentence), bbox_from_span(span)
                    )
                    and sentence is not span.sentence
                ):
                    for ngram in tokens_to_ngrams(
                        getattr(sentence, attrib), n_min=n_min, n_max=n_max, lower=lower
                    ):
                        yield ngram
            else:
                for ts in ngrams_space.apply(sentence):
                    if bbox_direction_aligned(
                        bbox_from_span(ts), bbox_from_span(span)
                    ) and not (
                        sentence == span.sentence and ts.get_span() in span.get_span()
                    ):
                        yield f(ts.get_span())


def get_vert_ngrams_left(c):
    """Not implemented."""
    # TODO
    return


def get_vert_ngrams_right(c):
    """Not implemented."""
    # TODO
    return


def get_vert_ngrams_center(c):
    """Not implemented."""
    # TODO
    return


def get_visual_header_ngrams(c, axis=None):
    """Not implemented."""
    # TODO
    return


def get_visual_distance(c, axis=None):
    """Not implemented."""
    # TODO
    return


# Default dimensions for 8.5" x 11"
DEFAULT_WIDTH = 612
DEFAULT_HEIGHT = 792


def get_page_vert_percentile(
    mention, page_width=DEFAULT_WIDTH, page_height=DEFAULT_HEIGHT
):
    """Return which percentile from the TOP in the page the Mention is located in.

    Percentile is calculated where the top of the page is 0.0, and the bottom
    of the page is 1.0. For example, a Mention in at the top 1/4 of the page
    will have a percentile of 0.25.

    Page width and height are based on pt values::

        Letter      612x792
        Tabloid     792x1224
        Ledger      1224x792
        Legal       612x1008
        Statement   396x612
        Executive   540x720
        A0          2384x3371
        A1          1685x2384
        A2          1190x1684
        A3          842x1190
        A4          595x842
        A4Small     595x842
        A5          420x595
        B4          729x1032
        B5          516x729
        Folio       612x936
        Quarto      610x780
        10x14       720x1008

    and should match the source documents. Letter size is used by default.

    Note that if a candidate is passed in, only the vertical percentil of its
    first Mention is returned.

    :param mention: The Mention to evaluate
    :param page_width: The width of the page. Default to Letter paper width.
    :param page_height: The heigh of the page. Default to Letter paper height.
    :rtype: float in [0.0, 1.0]
    """
    span = _to_span(mention)
    return bbox_from_span(span).top / page_height


def get_page_horz_percentile(
    mention, page_width=DEFAULT_WIDTH, page_height=DEFAULT_HEIGHT
):
    """Return which percentile from the LEFT in the page the Mention is located in.

    Percentile is calculated where the left of the page is 0.0, and the right
    of the page is 1.0.

    Page width and height are based on pt values::

        Letter      612x792
        Tabloid     792x1224
        Ledger      1224x792
        Legal       612x1008
        Statement   396x612
        Executive   540x720
        A0          2384x3371
        A1          1685x2384
        A2          1190x1684
        A3          842x1190
        A4          595x842
        A4Small     595x842
        A5          420x595
        B4          729x1032
        B5          516x729
        Folio       612x936
        Quarto      610x780
        10x14       720x1008

    and should match the source documents. Letter size is used by default.

    Note that if a candidate is passed in, only the vertical percentile of its
    first Mention is returned.

    :param c: The Mention to evaluate
    :param page_width: The width of the page. Default to Letter paper width.
    :param page_height: The heigh of the page. Default to Letter paper height.
    :rtype: float in [0.0, 1.0]
    """
    span = _to_span(mention)
    return bbox_from_span(span).left / page_width


def _assign_alignment_features(sentences_by_key, align_type):
    for key, sentences in sentences_by_key.items():
        if len(sentences) == 1:
            continue
        context_lemmas = set()
        for p in sentences:
            p._aligned_lemmas.update(context_lemmas)
            # update lemma context for upcoming sentences in the group
            if len(p.lemmas) < 7:
                new_lemmas = [lemma.lower() for lemma in p.lemmas if lemma.isalpha()]
                context_lemmas.update(new_lemmas)
                context_lemmas.update(align_type + lemma for lemma in new_lemmas)


def _preprocess_visual_features(doc):
    if hasattr(doc, "_visual_features"):
        return
    # cache flag
    doc._visual_features = True

    sentence_by_page = defaultdict(list)
    for sentence in doc.sentences:
        sentence_by_page[sentence.page[0]].append(sentence)
        sentence._aligned_lemmas = set()

    for page, sentences in sentence_by_page.items():
        # process per page alignments
        yc_aligned = defaultdict(list)
        x0_aligned = defaultdict(list)
        xc_aligned = defaultdict(list)
        x1_aligned = defaultdict(list)
        for sentence in sentences:
            sentence.bbox = bbox_from_sentence(sentence)
            sentence.yc = (sentence.bbox.top + sentence.bbox.bottom) / 2
            sentence.x0 = sentence.bbox.left
            sentence.x1 = sentence.bbox.right
            sentence.xc = (sentence.x0 + sentence.x1) / 2
            # index current sentence by different alignment keys
            yc_aligned[sentence.yc].append(sentence)
            x0_aligned[sentence.x0].append(sentence)
            x1_aligned[sentence.x1].append(sentence)
            xc_aligned[sentence.xc].append(sentence)
        for l in yc_aligned.values():
            l.sort(key=lambda p: p.xc)
        for l in x0_aligned.values():
            l.sort(key=lambda p: p.yc)
        for l in x1_aligned.values():
            l.sort(key=lambda p: p.yc)
        for l in xc_aligned.values():
            l.sort(key=lambda p: p.yc)
        _assign_alignment_features(yc_aligned, "Y_")
        _assign_alignment_features(x0_aligned, "LEFT_")
        _assign_alignment_features(x1_aligned, "RIGHT_")
        _assign_alignment_features(xc_aligned, "CENTER_")


def get_visual_aligned_lemmas(mention):
    """Return a generator of the lemmas aligned visually with the Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention to evaluate.
    :rtype: a *generator* of lemmas
    """
    spans = _to_spans(mention)
    for span in spans:
        sentence = span.sentence
        doc = sentence.document
        # cache features for the entire document
        _preprocess_visual_features(doc)

        for aligned_lemma in sentence._aligned_lemmas:
            yield aligned_lemma


def get_aligned_lemmas(mention):
    """Return a set of the lemmas aligned visually with the Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention to evaluate.
    :rtype: a set of lemmas
    """
    return set(get_visual_aligned_lemmas(mention))



from fonduer.utils.data_model_utils.structural import (
    common_ancestor,
    get_ancestor_class_names,
    get_ancestor_id_names,
    get_ancestor_tag_names,
    get_attributes,
    get_next_sibling_tags,
    get_parent_tag,
    get_prev_sibling_tags,
    get_tag,
    lowest_common_ancestor_depth,
)
from fonduer.utils.data_model_utils.tabular import (
    get_aligned_ngrams,
    get_cell_ngrams,
    get_col_ngrams,
    get_head_ngrams,
    get_max_col_num,
    get_min_col_num,
    get_neighbor_cell_ngrams,
    get_neighbor_sentence_ngrams,
    get_row_ngrams,
    get_sentence_ngrams,
    is_tabular_aligned,
    same_cell,
    same_col,
    same_document,
    same_row,
    same_sentence,
    same_table,
)
from fonduer.utils.data_model_utils.textual import (
    get_between_ngrams,
    get_left_ngrams,
    get_right_ngrams,
)
from fonduer.utils.data_model_utils.utils import get_matches, is_superset, overlap
from fonduer.utils.data_model_utils.visual import (
    get_aligned_lemmas,
    get_horz_ngrams,
    get_page,
    get_page_horz_percentile,
    get_page_vert_percentile,
    get_vert_ngrams,
    get_visual_aligned_lemmas,
    is_horz_aligned,
    is_vert_aligned,
    is_vert_aligned_center,
    is_vert_aligned_left,
    is_vert_aligned_right,
    same_page,
)

__all__ = [
    "common_ancestor",
    "get_aligned_lemmas",
    "get_aligned_ngrams",
    "get_ancestor_class_names",
    "get_ancestor_id_names",
    "get_ancestor_tag_names",
    "get_attributes",
    "get_between_ngrams",
    "get_cell_ngrams",
    "get_col_ngrams",
    "get_head_ngrams",
    "get_horz_ngrams",
    "get_left_ngrams",
    "get_matches",
    "get_max_col_num",
    "get_min_col_num",
    "get_neighbor_cell_ngrams",
    "get_neighbor_sentence_ngrams",
    "get_next_sibling_tags",
    "get_page",
    "get_page_horz_percentile",
    "get_page_vert_percentile",
    "get_parent_tag",
    "get_prev_sibling_tags",
    "get_right_ngrams",
    "get_row_ngrams",
    "get_sentence_ngrams",
    "get_tag",
    "get_vert_ngrams",
    "get_visual_aligned_lemmas",
    "is_horz_aligned",
    "is_superset",
    "is_tabular_aligned",
    "is_vert_aligned",
    "is_vert_aligned_center",
    "is_vert_aligned_left",
    "is_vert_aligned_right",
    "lowest_common_ancestor_depth",
    "overlap",
    "same_cell",
    "same_col",
    "same_document",
    "same_page",
    "same_row",
    "same_sentence",
    "same_table",
]



import logging
import os
import re
import shutil
import subprocess
from builtins import object, range, str, zip
from collections import OrderedDict, defaultdict

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from editdistance import eval as editdist  # Alternative library: python-levenshtein


class VisualLinker(object):
    def __init__(self, time=False, verbose=False):
        self.logger = logging.getLogger(__name__)
        self.pdf_file = None
        self.verbose = verbose
        self.time = time
        self.coordinate_map = None
        self.pdf_word_list = None
        self.html_word_list = None
        self.links = None
        self.pdf_dim = None
        delimiters = (
            u"([\(\)\,\?\u2212\u201C\u201D\u2018\u2019\u00B0\*']|(?<!http):|\.$|\.\.\.)"
        )
        self.separators = re.compile(delimiters)

        # Check if poppler-utils is installed AND the version is 0.36.0 or above
        if shutil.which("pdfinfo") is None or shutil.which("pdftotext") is None:
            raise RuntimeError("poppler-utils is not installed or they are not in PATH")
        version = subprocess.check_output(
            "pdfinfo -v", shell=True, stderr=subprocess.STDOUT, universal_newlines=True
        )
        m = re.search(r"\d\.\d{2}\.\d", version)
        if int(m.group(0).replace(".", "")) < 360:
            raise RuntimeError(
                "Installed poppler-utils's version is %s, but should be 0.36.0 or above"
                % m.group(0)
            )

    def parse_visual(self, document_name, sentences, pdf_path):
        self.sentences = sentences
        self.pdf_file = (
            pdf_path if os.path.isfile(pdf_path) else pdf_path + document_name + ".pdf"
        )
        if not os.path.isfile(self.pdf_file):
            self.pdf_file = self.pdf_file[:-3] + "PDF"
        try:
            self.extract_pdf_words()
        except RuntimeError as e:
            self.logger.exception(e)
            return
        self.extract_html_words()
        self.link_lists(search_max=200)
        for sentence in self.update_coordinates():
            yield sentence

    def extract_pdf_words(self):
        self.logger.debug(
            "pdfinfo '{}' | grep -a Pages | sed 's/[^0-9]*//'".format(self.pdf_file)
        )
        num_pages = subprocess.check_output(
            "pdfinfo '{}' | grep -a Pages | sed 's/[^0-9]*//'".format(self.pdf_file),
            shell=True,
        )
        pdf_word_list = []
        coordinate_map = {}
        for i in range(1, int(num_pages) + 1):
            self.logger.debug(
                "pdftotext -f {} -l {} -bbox-layout '{}' -".format(
                    str(i), str(i), self.pdf_file
                )
            )
            html_content = subprocess.check_output(
                "pdftotext -f {} -l {} -bbox-layout '{}' -".format(
                    str(i), str(i), self.pdf_file
                ),
                shell=True,
            )
            soup = BeautifulSoup(html_content, "html.parser")
            pages = soup.find_all("page")
            pdf_word_list_i, coordinate_map_i = self._coordinates_from_HTML(pages[0], i)
            pdf_word_list += pdf_word_list_i
            # update coordinate map
            coordinate_map.update(coordinate_map_i)
        self.pdf_word_list = pdf_word_list
        self.coordinate_map = coordinate_map
        if len(self.pdf_word_list) == 0:
            raise RuntimeError(
                "Words could not be extracted from PDF: %s" % self.pdf_file
            )
        # take last page dimensions
        page_width, page_height = (
            int(float(pages[0].get("width"))),
            int(float(pages[0].get("height"))),
        )
        self.pdf_dim = (page_width, page_height)
        if self.verbose:
            self.logger.info("Extracted {} pdf words".format(len(self.pdf_word_list)))

    def _coordinates_from_HTML(self, page, page_num):
        pdf_word_list = []
        coordinate_map = {}
        block_coordinates = {}
        blocks = page.find_all("block")
        i = 0  # counter for word_id in page_num
        for block in blocks:
            x_min_block = int(float(block.get("xmin")))
            y_min_block = int(float(block.get("ymin")))
            lines = block.find_all("line")
            for line in lines:
                y_min_line = int(float(line.get("ymin")))
                y_max_line = int(float(line.get("ymax")))
                words = line.find_all("word")
                for word in words:
                    xmin = int(float(word.get("xmin")))
                    xmax = int(float(word.get("xmax")))
                    for content in self.separators.split(word.getText()):
                        if len(content) > 0:  # Ignore empty characters
                            word_id = (page_num, i)
                            pdf_word_list.append((word_id, content))
                            coordinate_map[word_id] = (
                                page_num,
                                y_min_line,
                                xmin,
                                y_max_line,
                                xmax,
                            )
                            block_coordinates[word_id] = (y_min_block, x_min_block)
                            i += 1
        # sort pdf_word_list by page, block top then block left, top, then left
        pdf_word_list = sorted(
            pdf_word_list,
            key=lambda word_id__: block_coordinates[word_id__[0]]
            + coordinate_map[word_id__[0]][1:3],
        )
        return pdf_word_list, coordinate_map

    def extract_html_words(self):
        html_word_list = []
        for sentence in self.sentences:
            for i, word in enumerate(sentence.words):
                html_word_list.append(((sentence.stable_id, i), word))
        self.html_word_list = html_word_list
        if self.verbose:
            self.logger.info("Extracted {} html words".format(len(self.html_word_list)))

    def link_lists(self, search_max=100, edit_cost=20, offset_cost=1):
        # NOTE: there are probably some inefficiencies here from rehashing words
        # multiple times, but we're not going to worry about that for now

        def link_exact(l, u):
            l, u, L, U = get_anchors(l, u)
            html_dict = defaultdict(list)
            pdf_dict = defaultdict(list)
            for i, (_, word) in enumerate(self.html_word_list[l:u]):
                if html_to_pdf[l + i] is None:
                    html_dict[word].append(l + i)
            for j, (_, word) in enumerate(self.pdf_word_list[L:U]):
                if pdf_to_html[L + j] is None:
                    pdf_dict[word].append(L + j)
            for word, html_list in list(html_dict.items()):
                pdf_list = pdf_dict[word]
                if len(html_list) == len(pdf_list):
                    for k in range(len(html_list)):
                        html_to_pdf[html_list[k]] = pdf_list[k]
                        pdf_to_html[pdf_list[k]] = html_list[k]

        def link_fuzzy(i):
            (_, word) = self.html_word_list[i]
            l = u = i
            l, u, L, U = get_anchors(l, u)
            offset = int(L + float(i - l) / (u - l) * (U - L))
            searchIndices = np.clip(offset + search_order, 0, M - 1)
            cost = [0] * search_max
            for j, k in enumerate(searchIndices):
                other = self.pdf_word_list[k][1]
                if (
                    word.startswith(other)
                    or word.endswith(other)
                    or other.startswith(word)
                    or other.endswith(word)
                ):
                    html_to_pdf[i] = k
                    return
                else:
                    cost[j] = int(editdist(word, other)) * edit_cost + j * offset_cost
            html_to_pdf[i] = searchIndices[np.argmin(cost)]
            return

        def get_anchors(l, u):
            while l >= 0 and html_to_pdf[l] is None:
                l -= 1
            while u < N and html_to_pdf[u] is None:
                u += 1
            if l < 0:
                l = 0
                L = 0
            else:
                L = html_to_pdf[l]
            if u >= N:
                u = N
                U = M
            else:
                U = html_to_pdf[u]
            return l, u, L, U

        def display_match_counts():
            matches = sum(
                [
                    html_to_pdf[i] is not None
                    and self.html_word_list[i][1]
                    == self.pdf_word_list[html_to_pdf[i]][1]
                    for i in range(len(self.html_word_list))
                ]
            )
            total = len(self.html_word_list)
            self.logger.info(
                "({:d}/{:d}) = {:.2f}".format(matches, total, matches / total)
            )
            return matches

        N = len(self.html_word_list)
        M = len(self.pdf_word_list)

        try:
            assert N > 0 and M > 0
        except Exception:
            self.logger.exception("N = {} and M = {} are invalid values.".format(N, M))

        html_to_pdf = [None] * N
        pdf_to_html = [None] * M
        search_radius = search_max // 2

        # first pass: global search for exact matches
        link_exact(0, N)
        if self.verbose:
            self.logger.debug("Global exact matching:")
            display_match_counts()

        # second pass: local search for exact matches
        for i in range(((N + 2) // search_radius) + 1):
            link_exact(
                max(0, i * search_radius - search_radius),
                min(N, i * search_radius + search_radius),
            )
        if self.verbose:
            self.logger.debug("Local exact matching:")
            display_match_counts()

        # third pass: local search for approximate matches
        search_order = np.array(
            [(-1) ** (i % 2) * (i // 2) for i in range(1, search_max + 1)]
        )
        for i in range(len(html_to_pdf)):
            if html_to_pdf[i] is None:
                link_fuzzy(i)
        if self.verbose:
            self.logger.debug("Local approximate matching:")
            display_match_counts()

        # convert list to dict
        matches = sum(
            [
                html_to_pdf[i] is not None
                and self.html_word_list[i][1] == self.pdf_word_list[html_to_pdf[i]][1]
                for i in range(len(self.html_word_list))
            ]
        )
        total = len(self.html_word_list)
        if self.verbose:
            self.logger.debug(
                "Linked {:d}/{:d} ({:.2f}) html words exactly".format(
                    matches, total, matches / total
                )
            )
        self.links = OrderedDict(
            (self.html_word_list[i][0], self.pdf_word_list[html_to_pdf[i]][0])
            for i in range(len(self.html_word_list))
        )

    def _calculate_offset(self, listA, listB, seedSize, maxOffset):
        wordsA = zip(*listA[:seedSize])[1]
        wordsB = zip(*listB[:maxOffset])[1]
        offsets = []
        for i in range(seedSize):
            try:
                offsets.append(wordsB.index(wordsA[i]) - i)
            except Exception as e:
                pass
        return int(np.median(offsets))

    def display_links(self, max_rows=100):
        html = []
        pdf = []
        j = []
        for i, l in enumerate(self.links):
            html.append(self.html_word_list[i][1])
            for k, b in enumerate(self.pdf_word_list):
                if b[0] == self.links[self.html_word_list[i][0]]:
                    pdf.append(b[1])
                    j.append(k)
                    break
        try:
            assert len(pdf) == len(html)
        except Exception:
            self.logger.exception("PDF and HTML are not the same length")

        total = 0
        match = 0
        for i, word in enumerate(html):
            total += 1
            if word == pdf[i]:
                match += 1
        self.logger.info((match, total, match / total))

        data = {
            # 'i': range(len(self.links)),
            "html": html,
            "pdf": pdf,
            "j": j,
        }
        pd.set_option("display.max_rows", max_rows)
        self.logger.info(pd.DataFrame(data, columns=["html", "pdf", "j"]))
        pd.reset_option("display.max_rows")

    def update_coordinates(self):
        for sentence in self.sentences:
            (page, top, left, bottom, right) = list(
                zip(
                    *[
                        self.coordinate_map[self.links[((sentence.stable_id), i)]]
                        for i in range(len(sentence.words))
                    ]
                )
            )
            sentence.page = list(page)
            sentence.top = list(top)
            sentence.left = list(left)
            sentence.bottom = list(bottom)
            sentence.right = list(right)
            yield sentence
        if self.verbose:
            self.logger.debug("Updated coordinates in database")



import re

import lxml.etree as et


class XMLTree:
    """A generic tree representation which takes XML as input.

    Includes subroutines for conversion to JSON & for visualization based on js
    form
    """

    def __init__(self, xml_root, words=None):
        """Calls subroutines to generate JSON form of XML input"""
        self.root = xml_root
        self.words = words

        # create a unique id for e.g. canvas id in notebook
        self.id = str(abs(hash(self.to_str())))

    def _to_json(self, root):
        js = {"attrib": dict(root.attrib), "children": []}
        for i, c in enumerate(root):
            js["children"].append(self._to_json(c))
        return js

    def to_json(self):
        return self._to_json(self.root)

    def to_str(self):
        return et.tostring(self.root)


def corenlp_to_xmltree(s, prune_root=True):
    """
    Transforms an object with CoreNLP dep_path and dep_parent attributes into
    an XMLTree. Will include elements of any array having the same dimensiion
    as dep_* as node attributes. Also adds special word_idx attribute
    corresponding to original sequence order in sentence.
    """
    # Convert input object to dictionary
    if not isinstance(s, dict):
        try:
            s = s.__dict__ if hasattr(s, "__dict__") else dict(s)
        except Exception as e:
            raise ValueError("Cannot convert input object to dict")

    # Use the dep_parents array as a guide: ensure it is present and a list of
    # ints
    if not ("dep_parents" in s and isinstance(s["dep_parents"], list)):
        raise ValueError(
            "Input CoreNLP object must have a 'dep_parents' attribute which is a list"
        )
    try:
        dep_parents = list(map(int, s["dep_parents"]))
    except Exception as e:
        raise ValueError("'dep_parents' attribute must be a list of ints")

    # Also ensure that we are using CoreNLP-native indexing
    # (root=0, 1-base word indexes)!
    b = min(dep_parents)
    if b != 0:
        dep_parents = list(map(lambda j: j - b, dep_parents))

    # Parse recursively
    root = corenlp_to_xmltree_sub(s, dep_parents, 0)

    # Often the return tree will have several roots, where one is the actual
    # root and the rest are just singletons not included in the dep tree
    # parse...
    # We optionally remove these singletons and then collapse the root if only
    # one child left.
    if prune_root:
        for c in root:
            if len(c) == 0:
                root.remove(c)
        if len(root) == 1:
            root = root.findall("./*")[0]
    return XMLTree(root, words=s["words"])


def scrub(s):
    return "".join(c for c in s if ord(c) < 128)


def corenlp_to_xmltree_sub(s, dep_parents, rid=0):
    i = rid - 1
    attrib = {}
    N = len(list(dep_parents))

    # Add all attributes that have the same shape as dep_parents
    if i >= 0:
        for k, v in list(
            filter(lambda t: isinstance(t[1], list) and len(t[1]) == N, s.items())
        ):
            if v[i] is not None:
                attrib[singular(k)] = (
                    scrub(v[i]).encode("ascii", "ignore")
                    if hasattr(v[i], "encode")
                    else str(v[i])
                )

        # Add word_idx if not present
        if "word_idx" not in attrib:
            attrib["word_idx"] = str(i)

    # Build tree recursively
    root = et.Element("node", attrib=attrib)
    for i, d in enumerate(dep_parents):
        if d == rid:
            root.append(corenlp_to_xmltree_sub(s, dep_parents, i + 1))
    return root


def singular(s):
    """Get singular form of word s (crudely)"""
    return re.sub(r"e?s$", "", s, flags=re.I)



import codecs
import os

from fonduer.parser.models import Document
from fonduer.parser.preprocessors.doc_preprocessor import DocPreprocessor


class TextDocPreprocessor(DocPreprocessor):
    """A generator which processes a text file or directory of text files into
    a set of Document objects.

    Assumes one document per file.

    :param encoding: file encoding to use (e.g. "utf-8").
    :type encoding: str
    :param path: filesystem path to file or directory to parse.
    :type path: str
    :param max_docs: the maximum number of ``Documents`` to produce.
    :type max_docs: int
    :rtype: A generator of ``Documents``.
    """

    def _parse_file(self, fp, file_name):
        with codecs.open(fp, encoding=self.encoding) as f:
            name = os.path.basename(fp).rsplit(".", 1)[0]
            stable_id = self._get_stable_id(name)
            doc = Document(
                name=name,
                stable_id=stable_id,
                meta={"file_name": file_name},
                text=f.read(),
            )
            yield doc



from sqlalchemy import Column, ForeignKey, Integer, String

from fonduer.parser.models.document import Document


class Webpage(Document):
    """A Webpage document context enhanced with additional metadata."""

    __tablename__ = "webpage"

    #: The unique id of the ``Webpage``.
    id = Column(
        Integer, ForeignKey("document.id", ondelete="CASCADE"), primary_key=True
    )
    #: The URL of the ``Webpage``.
    url = Column(String)
    #: The host of the ``Webpage``.
    host = Column(String)

    #: The type of the ``Webpage``.
    page_type = Column(String)

    #: The raw content of the ``Webpage``.
    raw_content = Column(String)

    #: The timestamp of when the ``Webpage`` was crawled.
    crawltime = Column(String)
    all = Column(String)

    # Polymorphism information for SQLAlchemy
    __mapper_args__ = {"polymorphic_identity": "webpage"}

    # Rest of class definition here
    def __repr__(self):
        return "Webpage(id: {}..., url: {}...)".format(self.name[:10], self.url[8:23])



"""
A recurrent neural network module.
"""

import torch
import torch.nn as nn


class RNN(nn.Module):
    """A recurrent neural network layer.

    :param num_classes: Number of classes.
    :type num_classes: int
    :param num_tokens: Size of embeddings.
    :type num_tokens: int
    :param emb_size: Dimension of embeddings.
    :type emb_size: int
    :param lstm_hidden: Size of LSTM hidden layer size.
    :type lstm_hidden: int
    :param num_layers: Number of recurrent layers.
    :type num_layers: int
    :param dropout: Dropout parameter of LSTM.
    :type dropout: float
    :param attention: Use attention or not.
    :type attention: bool
    :param bidirectional: Use bidirectional LSTM or not.
    :type bidirectional: bool
    :param use_cuda: Use use_cuda or not.
    :type use_cuda: bool
    """

    def __init__(
        self,
        num_classes,
        num_tokens,
        emb_size,
        lstm_hidden,
        num_layers=1,
        dropout=0.0,
        attention=True,
        bidirectional=True,
        use_cuda=False,
    ):

        super(RNN, self).__init__()

        self.num_tokens = num_tokens
        self.emb_size = emb_size
        self.lstm_hidden = lstm_hidden
        self.bidirectional = bidirectional
        self.attention = attention
        self.num_layers = num_layers
        self.dropout = dropout
        self.use_cuda = use_cuda
        # Number of dimensions of output.  means no final linear layer.
        self.num_classes = num_classes

        if self.num_classes > 0:
            self.final_linear = True
        else:
            self.final_linear = False

        self.drop = nn.Dropout(dropout)
        self.lookup = nn.Embedding(num_tokens, emb_size, padding_idx=0)

        b = 2 if self.bidirectional else 1

        self.word_lstm = nn.LSTM(
            self.emb_size,
            self.lstm_hidden,
            batch_first=True,
            num_layers=self.num_layers,
            dropout=self.dropout if self.num_layers > 1 else 0.0,
            bidirectional=self.bidirectional,
        )

        if attention:
            self.attn_linear_w_1 = nn.Linear(
                b * lstm_hidden, b * lstm_hidden, bias=True
            )
            self.attn_linear_w_2 = nn.Linear(b * lstm_hidden, 1, bias=False)

        if self.final_linear:
            self.linear = nn.Linear(b * lstm_hidden, num_classes)

    def forward(self, x, x_mask, state_word):
        """Forward function.

        :param x: Input sequence tensor.
        :type x: torch.Tensor of shape (batch_size * length)
        :param x_mask: Use use_cuda or not.
        :type x_mask: torch.Tensor of shape (batch_size * length)
        :param state_word: Initial state of LSTM.
        :type state_word: torch.Tensor (see init_hidden() for more information)
        :return: Output of LSTM layer, either after mean pooling or attention.
        :rtype: torch.Tensor with shape (batch_size, num_directions * hidden_size)
            if num_classes > 0 otherwise with shape (batch_size, num_classes)
        """

        x_emb = self.drop(self.lookup(x))
        output_word, state_word = self.word_lstm(x_emb, state_word)
        output_word = self.drop(output_word)
        if self.attention:
            """
            An attention layer where the attention weight is
            a = T' . tanh(Wx + b)
            where x is the input, b is the bias.
            """
            word_squish = torch.tanh(self.attn_linear_w_1(output_word))
            word_attn = self.attn_linear_w_2(word_squish)
            word_attn.data.masked_fill_(x_mask.data.unsqueeze(dim=2), float("-inf"))
            word_attn_norm = torch.sigmoid(word_attn.squeeze(2))
            word_attn_vectors = torch.bmm(
                output_word.transpose(1, 2), word_attn_norm.unsqueeze(2)
            ).squeeze(2)
            output = (
                self.linear(word_attn_vectors)
                if self.final_linear
                else word_attn_vectors
            )
        else:
            """
            Mean pooling
            """
            x_lens = x_mask.data.eq(0).long().sum(dim=1)
            if self.use_cuda:
                weights = torch.ones(x.size()).cuda() / x_lens.unsqueeze(1).float()
            else:
                weights = torch.ones(x.size()) / x_lens.unsqueeze(1).float()
            weights.data.masked_fill_(x_mask.data.unsqueeze(dim=2), 0.0)
            word_vectors = torch.bmm(
                output_word.transpose(1, 2), weights.unsqueeze(2)
            ).squeeze(2)
            output = self.linear(word_vectors) if self.final_linear else word_vectors

        return output

    def init_hidden(self, batch_size):
        """Initiate the initial state.

        :param batch_size: batch size.
        :type batch_size: int
        :return: Initial state of LSTM
        :rtype: pair of torch.Tensors of shape (num_layers * num_directions,
            batch_size, hidden_size)
        """

        b = 2 if self.bidirectional else 1
        if self.use_cuda:
            return (
                torch.zeros(self.num_layers * b, batch_size, self.lstm_hidden).cuda(),
                torch.zeros(self.num_layers * b, batch_size, self.lstm_hidden).cuda(),
            )
        else:
            return (
                torch.zeros(self.num_layers * b, batch_size, self.lstm_hidden),
                torch.zeros(self.num_layers * b, batch_size, self.lstm_hidden),
            )



import logging
import os

import yaml

MAX_CONFIG_SEARCH_DEPTH = 25  # Max num of parent directories to look for config
logger = logging.getLogger(__name__)

default = {
    "featurization": {
        "content": {
            "window_feature": {"size": 3, "combinations": True, "isolated": True},
            "word_feature": {"window": 7},
        },
        "table": {
            "unary_features": {
                "attrib": ["words"],
                "get_cell_ngrams": {"max": 2},
                "get_head_ngrams": {"max": 2},
                "get_row_ngrams": {"max": 2},
                "get_col_ngrams": {"max": 2},
            },
            "binary_features": {
                "min_row_diff": {"absolute": False},
                "min_col_diff": {"absolute": False},
            },
        },
    },
    "learning": {
        "LSTM": {
            "emb_dim": 100,
            "hidden_dim": 100,
            "attention": True,
            "dropout": 0.1,
            "bidirectional": True,
            "host_device": "CPU",
            "max_sentence_length": 100,
        },
        "LogisticRegression": {"bias": False},
        "SparseLogisticRegression": {"bias": False},
    },
}


def _merge(x, y):
    """Merge two nested dictionaries. Overwrite values in x with values in y."""
    merged = {**x, **y}

    xkeys = x.keys()

    for key in xkeys:
        if isinstance(x[key], dict) and key in y:
            merged[key] = _merge(x[key], y[key])

    return merged


def get_config(path=os.getcwd()):
    """Search for settings file in root of project and its parents."""
    config = default
    tries = 0
    current_dir = path
    while current_dir and tries < MAX_CONFIG_SEARCH_DEPTH:
        potential_path = os.path.join(current_dir, ".fonduer-config.yaml")
        if os.path.exists(potential_path):
            with open(potential_path, "r") as f:
                config = _merge(config, yaml.safe_load(f))
            logger.debug("Loading Fonduer config from {}.".format(potential_path))
            break

        new_dir = os.path.split(current_dir)[0]
        if current_dir == new_dir:
            logger.debug("Unable to find config file. Using defaults.")
            break
        current_dir = new_dir
        tries += 1

    return config



############################
# Tabular modality utilities
############################

from builtins import range
from itertools import chain

from fonduer.parser.models import Sentence
from fonduer.utils.data_model_utils.textual import get_left_ngrams, get_right_ngrams
from fonduer.utils.data_model_utils.utils import _to_span, _to_spans
from fonduer.utils.utils import tokens_to_ngrams
from fonduer.utils.utils_table import (
    is_axis_aligned,
    is_col_aligned,
    is_row_aligned,
    min_col_diff,
    min_row_diff,
)


def same_document(c):
    """Return True if all Mentions in the given candidate are from the same Document.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return all(
        _to_span(c[i]).sentence.document is not None
        and _to_span(c[i]).sentence.document == _to_span(c[0]).sentence.document
        for i in range(len(c))
    )


def same_table(c):
    """Return True if all Mentions in the given candidate are from the same Table.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return all(
        _to_span(c[i]).sentence.is_tabular()
        and _to_span(c[i]).sentence.table == _to_span(c[0]).sentence.table
        for i in range(len(c))
    )


def same_row(c):
    """Return True if all Mentions in the given candidate are from the same Row.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return same_table(c) and all(
        is_row_aligned(_to_span(c[i]).sentence, _to_span(c[0]).sentence)
        for i in range(len(c))
    )


def same_col(c):
    """Return True if all Mentions in the given candidate are from the same Col.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return same_table(c) and all(
        is_col_aligned(_to_span(c[i]).sentence, _to_span(c[0]).sentence)
        for i in range(len(c))
    )


def is_tabular_aligned(c):
    """Return True if all Mentions in the given candidate are from the same Row or Col.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return same_table(c) and (
        is_col_aligned(_to_span(c[i]).sentence, _to_span(c[0]).sentence)
        or is_row_aligned(_to_span(c[i]).sentence, _to_span(c[0]).sentence)
        for i in range(len(c))
    )


def same_cell(c):
    """Return True if all Mentions in the given candidate are from the same Cell.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return all(
        _to_span(c[i]).sentence.cell is not None
        and _to_span(c[i]).sentence.cell == _to_span(c[0]).sentence.cell
        for i in range(len(c))
    )


def same_sentence(c):
    """Return True if all Mentions in the given candidate are from the same Sentence.

    :param c: The candidate whose Mentions are being compared
    :rtype: boolean
    """
    return all(
        _to_span(c[i]).sentence is not None
        and _to_span(c[i]).sentence == _to_span(c[0]).sentence
        for i in range(len(c))
    )


def get_max_col_num(mention):
    """Return the largest column number that a Mention occupies.

    :param mention: The Mention to evaluate. If a candidate is given, default
        to its last Mention.
    :rtype: integer or None
    """
    span = _to_span(mention, idx=-1)
    if span.sentence.is_tabular():
        return span.sentence.cell.col_end
    else:
        return None


def get_min_col_num(mention):
    """Return the lowest column number that a Mention occupies.

    :param mention: The Mention to evaluate. If a candidate is given, default
        to its first Mention.
    :rtype: integer or None
    """
    span = _to_span(mention)
    if span.sentence.is_tabular():
        return span.sentence.cell.col_start
    else:
        return None


def get_sentence_ngrams(mention, attrib="words", n_min=1, n_max=1, lower=True):
    """Get the ngrams that are in the Sentence of the given Mention, not including itself.

    Note that if a candidate is passed in, all of its Mentions will be
    searched.

    :param mention: The Mention whose Sentence is being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in get_left_ngrams(
            span, window=100, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower
        ):
            yield ngram
        for ngram in get_right_ngrams(
            span, window=100, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower
        ):
            yield ngram


def get_neighbor_sentence_ngrams(
    mention, d=1, attrib="words", n_min=1, n_max=1, lower=True
):
    """Get the ngrams that are in the neighoring Sentences of the given Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention whose neighbor Sentences are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in chain.from_iterable(
            [
                tokens_to_ngrams(
                    getattr(sentence, attrib), n_min=n_min, n_max=n_max, lower=lower
                )
                for sentence in span.sentence.document.sentences
                if abs(sentence.sentence_num - span.sentence.sentence_num) <= d
                and sentence != span.sentence
            ]
        ):
            yield ngram


def get_cell_ngrams(mention, attrib="words", n_min=1, n_max=1, lower=True):
    """Get the ngrams that are in the Cell of the given mention, not including itself.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention whose Cell is being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in get_sentence_ngrams(
            span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower
        ):
            yield ngram
        if isinstance(span.sentence, Sentence) and span.sentence.cell is not None:
            for ngram in chain.from_iterable(
                [
                    tokens_to_ngrams(
                        getattr(sentence, attrib), n_min=n_min, n_max=n_max, lower=lower
                    )
                    for sentence in span.sentence.cell.sentences
                    if sentence != span.sentence
                ]
            ):
                yield ngram


def get_neighbor_cell_ngrams(
    mention, dist=1, directions=False, attrib="words", n_min=1, n_max=1, lower=True
):
    """
    Get the ngrams from all Cells that are within a given Cell distance in one
    direction from the given Mention.

    Note that if a candidate is passed in, all of its Mentions will be
    searched. If `directions=True``, each ngram will be returned with a
    direction in {'UP', 'DOWN', 'LEFT', 'RIGHT'}.

    :param mention: The Mention whose neighbor Cells are being searched
    :param dist: The Cell distance within which a neighbor Cell must be to be
        considered
    :param directions: A Boolean expressing whether or not to return the
        direction of each ngram
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams (or (ngram, direction) tuples if directions=True)
    """
    # TODO: Fix this to be more efficient (optimize with SQL query)
    spans = _to_spans(mention)
    for span in spans:
        for ngram in get_sentence_ngrams(
            span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower
        ):
            yield ngram
        if isinstance(span.sentence, Sentence) and span.sentence.cell is not None:
            root_cell = span.sentence.cell
            for sentence in chain.from_iterable(
                [
                    _get_aligned_sentences(root_cell, "row"),
                    _get_aligned_sentences(root_cell, "col"),
                ]
            ):
                row_diff = min_row_diff(sentence, root_cell, absolute=False)
                col_diff = min_col_diff(sentence, root_cell, absolute=False)
                if (
                    (row_diff or col_diff)
                    and not (row_diff and col_diff)
                    and abs(row_diff) + abs(col_diff) <= dist
                ):
                    if directions:
                        direction = ""
                        if col_diff == 0:
                            if 0 < row_diff and row_diff <= dist:
                                direction = "UP"
                            elif 0 > row_diff and row_diff >= -dist:
                                direction = "DOWN"
                        elif row_diff == 0:
                            if 0 < col_diff and col_diff <= dist:
                                direction = "RIGHT"
                            elif 0 > col_diff and col_diff >= -dist:
                                direction = "LEFT"
                        for ngram in tokens_to_ngrams(
                            getattr(sentence, attrib),
                            n_min=n_min,
                            n_max=n_max,
                            lower=lower,
                        ):
                            yield (ngram, direction)
                    else:
                        for ngram in tokens_to_ngrams(
                            getattr(sentence, attrib),
                            n_min=n_min,
                            n_max=n_max,
                            lower=lower,
                        ):
                            yield ngram


def get_row_ngrams(
    mention, attrib="words", n_min=1, n_max=1, spread=[0, 0], lower=True
):
    """Get the ngrams from all Cells that are in the same row as the given Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention whose row Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in _get_axis_ngrams(
            span,
            axis="row",
            attrib=attrib,
            n_min=n_min,
            n_max=n_max,
            spread=spread,
            lower=lower,
        ):
            yield ngram


def get_col_ngrams(
    mention, attrib="words", n_min=1, n_max=1, spread=[0, 0], lower=True
):
    """Get the ngrams from all Cells that are in the same column as the given Mention.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention whose column Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in _get_axis_ngrams(
            span,
            axis="col",
            attrib=attrib,
            n_min=n_min,
            n_max=n_max,
            spread=spread,
            lower=lower,
        ):
            yield ngram


def get_aligned_ngrams(
    mention, attrib="words", n_min=1, n_max=1, spread=[0, 0], lower=True
):
    """Get the ngrams from all Cells in the same row or column as the given Mention.

    Note that if a candidate is passed in, all of its Mentions will be
    searched.

    :param mention: The Mention whose row and column Cells are being searched
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    for span in spans:
        for ngram in get_row_ngrams(
            span, attrib=attrib, n_min=n_min, n_max=n_max, spread=spread, lower=lower
        ):
            yield ngram
        for ngram in get_col_ngrams(
            span, attrib=attrib, n_min=n_min, n_max=n_max, spread=spread, lower=lower
        ):
            yield ngram


def get_head_ngrams(mention, axis=None, attrib="words", n_min=1, n_max=1, lower=True):
    """Get the ngrams from the cell in the head of the row or column.

    More specifically, this returns the ngrams in the leftmost cell in a row and/or the
    ngrams in the topmost cell in the column, depending on the axis parameter.

    Note that if a candidate is passed in, all of its Mentions will be searched.

    :param mention: The Mention whose head Cells are being returned
    :param axis: Which axis {'row', 'col'} to search. If None, then both row
        and col are searched.
    :param attrib: The token attribute type (e.g. words, lemmas, poses)
    :param n_min: The minimum n of the ngrams that should be returned
    :param n_max: The maximum n of the ngrams that should be returned
    :param lower: If True, all ngrams will be returned in lower case
    :rtype: a *generator* of ngrams
    """
    spans = _to_spans(mention)
    axes = [axis] if axis else ["row", "col"]
    for span in spans:
        if not span.sentence.cell:
            return
        else:
            for axis in axes:
                if getattr(span.sentence, _other_axis(axis) + "_start") == 0:
                    return
                for sentence in getattr(
                    _get_head_cell(span.sentence.cell, axis), "sentences", []
                ):
                    for ngram in tokens_to_ngrams(
                        getattr(sentence, attrib), n_min=n_min, n_max=n_max, lower=lower
                    ):
                        yield ngram


def _get_head_cell(root_cell, axis):
    other_axis = "row" if axis == "col" else "col"
    aligned_cells = _get_aligned_cells(root_cell, axis)
    return (
        sorted(aligned_cells, key=lambda x: getattr(x, other_axis + "_start"))[0]
        if aligned_cells
        else []
    )


def _get_axis_ngrams(
    mention, axis, attrib="words", n_min=1, n_max=1, spread=[0, 0], lower=True
):
    span = _to_span(mention)
    for ngram in get_sentence_ngrams(
        span, attrib=attrib, n_min=n_min, n_max=n_max, lower=lower
    ):
        yield ngram
    if span.sentence.cell is not None:
        for sentence in _get_aligned_sentences(span.sentence, axis, spread=spread):
            for ngram in tokens_to_ngrams(
                getattr(sentence, attrib), n_min=n_min, n_max=n_max, lower=lower
            ):
                yield ngram


def _get_aligned_cells(root_cell, axis):
    aligned_cells = [
        cell
        for cell in root_cell.table.cells
        if is_axis_aligned(root_cell, cell, axis=axis) and cell != root_cell
    ]
    return aligned_cells


def _get_aligned_sentences(root_sentence, axis, spread=[0, 0]):
    return [
        sentence
        for cell in root_sentence.table.cells
        if is_axis_aligned(root_sentence, cell, axis=axis, spread=spread)
        for sentence in cell.sentences
        if sentence != root_sentence
    ]


def _other_axis(axis):
    return "row" if axis == "col" else "col"



###############################
# Structural modality utilities
###############################

import functools
from builtins import str

import numpy as np
from lxml import etree
from lxml.html import fromstring

from fonduer.utils.data_model_utils.utils import _to_span


def get_tag(mention):
    """Return the HTML tag of the Mention.

    If a candidate is passed in, only the tag of its first Mention is returned.

    These may be tags such as 'p', 'h2', 'table', 'div', etc.
    :param mention: The Mention to evaluate
    :rtype: string
    """
    span = _to_span(mention)
    return str(span.sentence.html_tag)


def get_attributes(mention):
    """Return the HTML attributes of the Mention.

    If a candidate is passed in, only the tag of its first Mention is returned.

    A sample outout of this function on a Mention in a paragraph tag is
    [u'style=padding-top: 8pt;padding-left: 20pt;text-indent: 0pt;text-align: left;']

    :param mention: The Mention to evaluate
    :rtype: list of strings representing HTML attributes
    """
    span = _to_span(mention)
    return span.sentence.html_attrs


@functools.lru_cache(maxsize=16)
def _get_etree_for_text(text):
    return etree.ElementTree(fromstring(text))


def _get_node(sentence):
    # Using caching to speed up retrieve process
    doc_etree = _get_etree_for_text(sentence.document.text)
    return doc_etree.xpath(sentence.xpath)[0]


def get_parent_tag(mention):
    """Return the HTML tag of the Mention's parent.

    These may be tags such as 'p', 'h2', 'table', 'div', etc.
    If a candidate is passed in, only the tag of its first Mention is returned.

    :param mention: The Mention to evaluate
    :rtype: string
    """
    span = _to_span(mention)
    i = _get_node(span.sentence)
    return str(i.getparent().tag) if i.getparent() is not None else None


def get_prev_sibling_tags(mention):
    """Return the HTML tag of the Mention's previous siblings.

    Previous siblings are Mentions which are at the same level in the HTML tree
    as the given mention, but are declared before the given mention. If a
    candidate is passed in, only the previous siblings of its first Mention are
    considered in the calculation.

    :param mention: The Mention to evaluate
    :rtype: list of strings
    """
    span = _to_span(mention)
    prev_sibling_tags = []
    i = _get_node(span.sentence)
    while i.getprevious() is not None:
        prev_sibling_tags.insert(0, str(i.getprevious().tag))
        i = i.getprevious()
    return prev_sibling_tags


def get_next_sibling_tags(mention):
    """Return the HTML tag of the Mention's next siblings.

    Next siblings are Mentions which are at the same level in the HTML tree as
    the given mention, but are declared after the given mention.
    If a candidate is passed in, only the next siblings of its last Mention
    are considered in the calculation.

    :param mention: The Mention to evaluate
    :rtype: list of strings
    """
    span = _to_span(mention)
    next_sibling_tags = []
    i = _get_node(span.sentence)
    while i.getnext() is not None:
        next_sibling_tags.append(str(i.getnext().tag))
        i = i.getnext()
    return next_sibling_tags


def get_ancestor_class_names(mention):
    """Return the HTML classes of the Mention's ancestors.

    If a candidate is passed in, only the ancestors of its first Mention are
    returned.

    :param mention: The Mention to evaluate
    :rtype: list of strings
    """
    span = _to_span(mention)
    class_names = []
    i = _get_node(span.sentence)
    while i is not None:
        class_names.insert(0, str(i.get("class")))
        i = i.getparent()
    return class_names


def get_ancestor_tag_names(mention):
    """Return the HTML tag of the Mention's ancestors.

    For example, ['html', 'body', 'p'].
    If a candidate is passed in, only the ancestors of its first Mention are returned.

    :param mention: The Mention to evaluate
    :rtype: list of strings
    """
    span = _to_span(mention)
    tag_names = []
    i = _get_node(span.sentence)
    while i is not None:
        tag_names.insert(0, str(i.tag))
        i = i.getparent()
    return tag_names


def get_ancestor_id_names(mention):
    """Return the HTML id's of the Mention's ancestors.

    If a candidate is passed in, only the ancestors of its first Mention are
    returned.

    :param mention: The Mention to evaluate
    :rtype: list of strings
    """
    span = _to_span(mention)
    id_names = []
    i = _get_node(span.sentence)
    while i is not None:
        id_names.insert(0, str(i.get("id")))
        i = i.getparent()
    return id_names


def common_ancestor(c):
    """Return the path to the root that is shared between a binary-Mention Candidate.

    In particular, this is the common path of HTML tags.

    :param c: The binary-Mention Candidate to evaluate
    :rtype: list of strings
    """
    span1 = _to_span(c[0])
    span2 = _to_span(c[1])
    ancestor1 = np.array(span1.sentence.xpath.split("/"))
    ancestor2 = np.array(span2.sentence.xpath.split("/"))
    min_len = min(ancestor1.size, ancestor2.size)
    return list(ancestor1[: np.argmin(ancestor1[:min_len] == ancestor2[:min_len])])


def lowest_common_ancestor_depth(c):
    """Return the minimum distance between a binary-Mention Candidate to their
    lowest common ancestor.

    For example, if the tree looked like this::

        html
        ├──<div> Mention 1 </div>
        ├──table
        │    ├──tr
        │    │  └──<th> Mention 2 </th>

    we return 1, the distance from Mention 1 to the html root. Smaller values
    indicate that two Mentions are close structurally, while larger values
    indicate that two Mentions are spread far apart structurally in the
    document.

    :param c: The binary-Mention Candidate to evaluate
    :rtype: integer
    """
    span1 = _to_span(c[0])
    span2 = _to_span(c[1])
    ancestor1 = np.array(span1.sentence.xpath.split("/"))
    ancestor2 = np.array(span2.sentence.xpath.split("/"))
    min_len = min(ancestor1.size, ancestor2.size)
    return min_len - np.argmin(ancestor1[:min_len] == ancestor2[:min_len])



from fonduer.features.feature_libs.content_features import get_content_feats
from fonduer.features.feature_libs.core_features import get_core_feats
from fonduer.features.feature_libs.structural_features import get_structural_feats
from fonduer.features.feature_libs.table_features import get_table_feats
from fonduer.features.feature_libs.visual_features import get_visual_feats


def get_all_feats(candidates):
    for candidate_id, feature, value in get_core_feats(candidates):
        yield candidate_id, feature, value
    for candidate_id, feature, value in get_content_feats(candidates):
        yield candidate_id, feature, value
    for candidate_id, feature, value in get_structural_feats(candidates):
        yield candidate_id, feature, value
    for candidate_id, feature, value in get_table_feats(candidates):
        yield candidate_id, feature, value
    for candidate_id, feature, value in get_visual_feats(candidates):
        yield candidate_id, feature, value


__all__ = [
    "get_all_feats",
    "get_content_feats",
    "get_core_feats",
    "get_structural_feats",
    "get_table_feats",
    "get_visual_feats",
]



from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import backref, relationship

from fonduer.candidates.models.temporarycontext import TemporaryContext
from fonduer.parser.models.context import Context


class TemporaryImage(TemporaryContext):
    """The TemporaryContext version of Image"""

    def __init__(self, figure):
        super(TemporaryImage, self).__init__()
        self.figure = figure  # The figure Context

    def __len__(self):
        return 1

    def __eq__(self, other):
        try:
            return (
                self.figure.position == other.figure.position
                and self.figure.url == other.figure.url
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return (
                self.figure.position != other.figure.position
                or self.figure.url != other.figure.url
            )
        except AttributeError:
            return True

    def __contains__(self, other_span):
        return self.__eq__(other_span)

    def __hash__(self):
        return hash(self.figure)

    def get_stable_id(self):
        """Return a stable id for the ``Image``."""
        return "%s::%s:%s" % (
            self.figure.document.id,
            self._get_polymorphic_identity(),
            self.figure.position,
        )

    def _get_table(self):
        return Image

    def _get_polymorphic_identity(self):
        return "image"

    def _get_insert_args(self):
        return {
            "document_id": self.figure.document.id,
            "position": self.figure.position,
            "url": self.figure.url,
        }

    def __repr__(self):
        return "{}(document={}, position={}, url={})".format(
            self.__class__.__name__,
            self.figure.document.name,
            self.figure.position,
            self.figure.url,
        )

    def _get_instance(self, **kwargs):
        return TemporaryImage(**kwargs)


class Image(Context, TemporaryImage):
    """A figure-based ``Candidate``.

    Note that this image should not be confused with the ``Figure`` context of
    the document model.
    """

    __tablename__ = "image"

    #: The unique id of the ``Image``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Document``.
    document_id = Column(Integer, ForeignKey("document.id", ondelete="CASCADE"))

    #: The position of the ``Image`` in the ``Document``.
    position = Column(Integer, nullable=False)

    #: The parent ``Document``.
    document = relationship(
        "Document",
        backref=backref("images", order_by=position, cascade="all, delete-orphan"),
        foreign_keys=document_id,
    )

    #: The url of the ``Image``.
    url = Column(String)

    __table_args__ = (UniqueConstraint(document_id, position),)

    __mapper_args__ = {"polymorphic_identity": "image"}

    def __repr__(self):
        return "Image(Doc: {}, Position: {}, Url: {})".format(
            self.document.name, self.position, self.url
        )

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import backref, relationship
from sqlalchemy.types import PickleType

from fonduer.candidates.models.temporarycontext import TemporaryContext
from fonduer.parser.models.context import Context
from fonduer.parser.models.utils import construct_stable_id


class TemporarySpan(TemporaryContext):
    """The TemporaryContext version of Span"""

    def __init__(self, sentence, char_start, char_end, meta=None):
        super(TemporarySpan, self).__init__()
        self.sentence = sentence  # The sentence Context of the Span
        self.char_end = char_end
        self.char_start = char_start
        self.meta = meta

    def __len__(self):
        return self.char_end - self.char_start + 1

    def __eq__(self, other):
        try:
            return (
                self.sentence == other.sentence
                and self.char_start == other.char_start
                and self.char_end == other.char_end
            )
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return (
                self.sentence != other.sentence
                or self.char_start != other.char_start
                or self.char_end != other.char_end
            )
        except AttributeError:
            return True

    def __hash__(self):
        return hash(self.sentence) + hash(self.char_start) + hash(self.char_end)

    def get_stable_id(self):
        """
        Return a stable id.

        :rtype: string
        """
        return construct_stable_id(
            self.sentence,
            self._get_polymorphic_identity(),
            self.char_start,
            self.char_end,
        )

    def _get_table(self):
        return Span

    def _get_polymorphic_identity(self):
        return "span"

    def _get_insert_args(self):
        return {
            "sentence_id": self.sentence.id,
            "char_start": self.char_start,
            "char_end": self.char_end,
            "meta": self.meta,
        }

    def get_word_start_index(self):
        """Get the index of the starting word of the span.

        :return: The word-index of the start of the span.
        :rtype: int
        """
        return self._char_to_word_index(self.char_start)

    def get_word_end_index(self):
        """Get the index of the ending word of the span.

        :return: The word-index of the last word of the span.
        :rtype: int
        """
        return self._char_to_word_index(self.char_end)

    def get_num_words(self):
        """Get the number of words in the span.

        :return: The number of words in the span (n of the ngrams).
        :rtype: int
        """
        return self.get_word_end_index() - self.get_word_start_index() + 1

    def _char_to_word_index(self, ci):
        """Return the index of the **word this char is in**.

        :param ci: The character-level index of the char.
        :type ci: int
        :return: The word-level index the char was in.
        :rtype: int
        """
        i = None
        for i, co in enumerate(self.sentence.char_offsets):
            if ci == co:
                return i
            elif ci < co:
                return i - 1
        return i

    def _word_to_char_index(self, wi):
        """Return the character-level index (offset) of the word's start.

        :param wi: The word-index.
        :type wi: int
        :return: The character-level index of the word's start.
        :rtype: int
        """
        return self.sentence.char_offsets[wi]

    def get_attrib_tokens(self, a="words"):
        """Get the tokens of sentence attribute *a*.

        Intuitively, like calling::

            span.a


        :param a: The attribute to get tokens for.
        :type a: str
        :return: The tokens of sentence attribute defined by *a* for the span.
        :rtype: list
        """
        return self.sentence.__getattribute__(a)[
            self.get_word_start_index() : self.get_word_end_index() + 1
        ]

    def get_attrib_span(self, a, sep=" "):
        """Get the span of sentence attribute *a*.

        Intuitively, like calling::

            sep.join(span.a)

        :param a: The attribute to get a span for.
        :type a: str
        :param sep: The separator to use for the join.
        :type sep: str
        :return: The joined tokens, or text if a="words".
        :rtype: str
        """
        # NOTE: Special behavior for words currently (due to correspondence
        # with char_offsets)
        if a == "words":
            return self.sentence.text[self.char_start : self.char_end + 1]
        else:
            return sep.join(self.get_attrib_tokens(a))

    def get_span(self):
        """Return the text of the ``Span``.

        :return: The text of the ``Span``.
        :rtype: str
        """
        return self.get_attrib_span("words")

    def __contains__(self, other_span):
        return (
            self.sentence == other_span.sentence
            and other_span.char_start >= self.char_start
            and other_span.char_end <= self.char_end
        )

    def __getitem__(self, key):
        """
        Slice operation returns a new candidate sliced according to **char index**

        Note that the slicing is w.r.t. the candidate range (not the abs.
        sentence char indexing).
        """
        if isinstance(key, slice):
            char_start = (
                self.char_start if key.start is None else self.char_start + key.start
            )
            if key.stop is None:
                char_end = self.char_end
            elif key.stop >= 0:
                char_end = self.char_start + key.stop - 1
            else:
                char_end = self.char_end + key.stop
            return self._get_instance(
                char_start=char_start, char_end=char_end, sentence=self.sentence
            )
        else:
            raise NotImplementedError()

    def __repr__(self):
        return '{}("{}", sentence={}, chars=[{},{}], words=[{},{}])'.format(
            self.__class__.__name__,
            self.get_span(),
            self.sentence.id,
            self.char_start,
            self.char_end,
            self.get_word_start_index(),
            self.get_word_end_index(),
        )

    def _get_instance(self, **kwargs):
        return TemporarySpan(**kwargs)


class Span(Context, TemporarySpan):
    """
    A span of chars, identified by Context ID and char-index start, end (inclusive).

    char_offsets are **relative to the Context start**
    """

    __tablename__ = "span"

    #: The unique id of the ``Span``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The id of the parent ``Sentence``.
    sentence_id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"))
    #: The parent ``Sentence``.
    sentence = relationship(
        "Context",
        backref=backref("spans", cascade="all, delete-orphan"),
        foreign_keys=sentence_id,
    )

    #: The starting character-index of the ``Span``.
    char_start = Column(Integer, nullable=False)
    #: The ending character-index of the ``Span`` (inclusive).
    char_end = Column(Integer, nullable=False)

    #: Pickled metadata about the ``ImplicitSpan``.
    meta = Column(PickleType)

    __table_args__ = (UniqueConstraint(sentence_id, char_start, char_end),)

    __mapper_args__ = {
        "polymorphic_identity": "span",
        "inherit_condition": (id == Context.id),
    }

    def _get_instance(self, **kwargs):
        return Span(**kwargs)

    # We redefine these to use default semantics, overriding the operators
    # inherited from TemporarySpan
    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return id(self)



import logging

from fonduer.candidates.models import Candidate
from fonduer.features.feature_libs import get_all_feats
from fonduer.features.models import Feature, FeatureKey
from fonduer.utils.udf import UDF, UDFRunner
from fonduer.utils.utils_udf import (
    ALL_SPLITS,
    add_keys,
    batch_upsert_records,
    get_cands_list_from_split,
    get_docs_from_split,
    get_mapping,
    get_sparse_matrix,
    get_sparse_matrix_keys,
)

logger = logging.getLogger(__name__)


class Featurizer(UDFRunner):
    """An operator to add Feature Annotations to Candidates.

    :param session: The database session to use.
    :param candidate_classes: A list of candidate_subclasses to featurize.
    :type candidate_classes: list
    :param parallelism: The number of processes to use in parallel. Default 1.
    :type parallelism: int
    """

    def __init__(self, session, candidate_classes, parallelism=1):
        """Initialize the Featurizer."""
        super(Featurizer, self).__init__(
            session,
            FeaturizerUDF,
            parallelism=parallelism,
            candidate_classes=candidate_classes,
        )
        self.candidate_classes = candidate_classes

    def update(self, docs=None, split=0, parallelism=None, progress_bar=True):
        """Update the features of the specified candidates.

        :param docs: If provided, apply features to all the candidates in these
            documents.
        :param split: If docs is None, apply features to the candidates in this
            particular split.
        :type split: int
        :param parallelism: How many threads to use for extraction. This will
            override the parallelism value used to initialize the Featurizer if
            it is provided.
        :type parallelism: int
        :param progress_bar: Whether or not to display a progress bar. The
            progress bar is measured per document.
        :type progress_bar: bool
        """
        self.apply(
            docs=docs,
            split=split,
            train=True,
            clear=False,
            parallelism=parallelism,
            progress_bar=progress_bar,
        )

    def apply(
        self,
        docs=None,
        split=0,
        train=False,
        clear=True,
        parallelism=None,
        progress_bar=True,
    ):
        """Apply features to the specified candidates.

        :param docs: If provided, apply features to all the candidates in these
            documents.
        :param split: If docs is None, apply features to the candidates in this
            particular split.
        :type split: int
        :param train: Whether or not to update the global key set of features
            and the features of candidates.
        :type train: bool
        :param clear: Whether or not to clear the features table before
            applying features.
        :type clear: bool
        :param parallelism: How many threads to use for extraction. This will
            override the parallelism value used to initialize the Featurizer if
            it is provided.
        :type parallelism: int
        :param progress_bar: Whether or not to display a progress bar. The
            progress bar is measured per document.
        :type progress_bar: bool
        """
        if docs:
            # Call apply on the specified docs for all splits
            split = ALL_SPLITS
            super(Featurizer, self).apply(
                docs,
                split=split,
                train=train,
                clear=clear,
                parallelism=parallelism,
                progress_bar=progress_bar,
            )
            # Needed to sync the bulk operations
            self.session.commit()
        else:
            # Only grab the docs containing candidates from the given split.
            split_docs = get_docs_from_split(
                self.session, self.candidate_classes, split
            )
            super(Featurizer, self).apply(
                split_docs,
                split=split,
                train=train,
                clear=clear,
                parallelism=parallelism,
                progress_bar=progress_bar,
            )
            # Needed to sync the bulk operations
            self.session.commit()

    def drop_keys(self, keys):
        """Drop the specified keys from FeatureKeys.

        :param keys: A list of FeatureKey names to delete.
        :type keys: list, tuple
        """
        # Make sure keys is iterable
        keys = keys if isinstance(keys, (list, tuple)) else [keys]

        # Remove the specified keys
        for key in keys:
            self.session.query(FeatureKey).filter(FeatureKey.name == key).delete()

    def get_keys(self):
        """Return a list of keys for the Features.

        :return: List of FeatureKeys.
        :rtype: list
        """
        return list(get_sparse_matrix_keys(self.session, FeatureKey))

    def clear(self, train=False, split=0):
        """Delete Features of each class from the database.

        :param train: Whether or not to clear the FeatureKeys
        :type train: bool
        :param split: Which split of candidates to clear features from.
        :type split: int
        """
        # Clear Features for the candidates in the split passed in.
        logger.info("Clearing Features (split {})".format(split))

        sub_query = (
            self.session.query(Candidate.id).filter(Candidate.split == split).subquery()
        )
        query = self.session.query(Feature).filter(Feature.candidate_id.in_(sub_query))
        query.delete(synchronize_session="fetch")

        # Delete all old annotation keys
        if train:
            logger.debug("Clearing all FeatureKey...")
            query = self.session.query(FeatureKey)
            query.delete(synchronize_session="fetch")

    def clear_all(self):
        """Delete all Features."""
        logger.info("Clearing ALL Features and FeatureKeys.")
        self.session.query(Feature).delete()
        self.session.query(FeatureKey).delete()

    def get_feature_matrices(self, cand_lists):
        """Load sparse matrix of Features for each candidate_class.

        :param cand_lists: The candidates to get features for.
        :type cand_lists: List of list of candidates.
        :return: An MxN sparse matrix where M are the candidates and N is the
            features.
        :rtype: csr_matrix
        """
        return get_sparse_matrix(self.session, FeatureKey, cand_lists)


class FeaturizerUDF(UDF):
    """UDF for performing candidate extraction."""

    def __init__(self, candidate_classes, **kwargs):
        """Initialize the FeaturizerUDF."""
        self.candidate_classes = (
            candidate_classes
            if isinstance(candidate_classes, (list, tuple))
            else [candidate_classes]
        )
        super(FeaturizerUDF, self).__init__(**kwargs)

    def apply(self, doc, split, train, **kwargs):
        """Extract candidates from the given Context.

        :param doc: A document to process.
        :param split: Which split to use.
        :param train: Whether or not to insert new FeatureKeys.
        """
        logger.debug("Document: {}".format(doc))

        # Get all the candidates in this doc that will be featurized
        cands_list = get_cands_list_from_split(
            self.session, self.candidate_classes, doc, split
        )

        feature_keys = set()
        for cands in cands_list:
            records = list(
                get_mapping(self.session, Feature, cands, get_all_feats, feature_keys)
            )
            batch_upsert_records(self.session, Feature, records)

        # Insert all Feature Keys
        if train:
            add_keys(self.session, FeatureKey, feature_keys)

        # This return + yield makes a completely empty generator
        return
        yield



from sqlalchemy import Column, ForeignKey, Integer, UniqueConstraint
from sqlalchemy.orm import backref, relationship

from fonduer.parser.models.context import Context


class Section(Context):
    """A Section Context in a Document.

    .. note:: Currently, each document simply has a single Section. Future
        parsing improvements can add better section recognition, such as the
        sections of an academic paper.
    """

    __tablename__ = "section"

    #: The unique id of the ``Section``.
    id = Column(Integer, ForeignKey("context.id", ondelete="CASCADE"), primary_key=True)

    #: The position of the ``Section`` in a ``Document``.
    position = Column(Integer, nullable=False)

    #: The id of the parent ``Document``.
    document_id = Column(Integer, ForeignKey("document.id", ondelete="CASCADE"))
    #: The parent ``Document``.
    document = relationship(
        "Document",
        backref=backref("sections", order_by=position, cascade="all, delete-orphan"),
        foreign_keys=document_id,
    )

    __mapper_args__ = {"polymorphic_identity": "section"}

    __table_args__ = (UniqueConstraint(document_id, position),)

    def __repr__(self):
        return "Section(Doc: {}, Pos: {})".format(self.document.name, self.position)

    def __gt__(self, other):
        # Allow sorting by comparing the string representations of each
        return self.__repr__() > other.__repr__()



from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import backref, relationship

from fonduer.meta import Meta
from fonduer.utils.utils import camel_to_under

# Grab pointer to global metadata
_meta = Meta.init()


class AnnotationKeyMixin(object):
    """Mixin class for defining annotation key tables.

    An AnnotationKey is the unique name associated with a set of Annotations,
    corresponding e.g. to a single labeling function or feature.
    """

    @declared_attr
    def __tablename__(cls):
        return camel_to_under(cls.__name__)

    @declared_attr
    def name(cls):
        """The name of the Key."""
        return Column(String, primary_key=True)

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint("name"),)

    def __repr__(self):
        return str(self.__class__.__name__) + " (" + str(self.name) + ")"


class AnnotationMixin(object):
    """Mixin class for defining annotation tables.

    An annotation is a value associated with a Candidate. Examples include
    labels, features, and predictions. New types of annotations can be defined
    by creating an annotation class and corresponding annotation, for example:

    .. code-block:: python

        from fonduer.utils.models import AnnotationMixin
        from fonduer.meta import Meta

        class NewAnnotation(AnnotationMixin, Meta.Base):
            values = Column(Float, nullable=False)

    The annotation class should include a Column attribute named values.
    """

    @declared_attr
    def __tablename__(cls):
        return camel_to_under(cls.__name__)

    # The key is the "name" or "type" of the Annotation- e.g. the name of a
    # feature, lf, or of a human annotator
    @declared_attr
    def keys(cls):
        """A list of strings of each Key name."""
        return Column(postgresql.ARRAY(String), nullable=False)

    # Every annotation is with respect to a candidate
    @declared_attr
    def candidate_id(cls):
        """The id of the ``Candidate`` being annotated."""
        return Column(
            "candidate_id",
            Integer,
            ForeignKey("candidate.id", ondelete="CASCADE"),
            primary_key=True,
        )

    @declared_attr
    def candidate(cls):
        """The ``Candidate``."""
        return relationship(
            "Candidate",
            backref=backref(
                camel_to_under(cls.__name__) + "s",
                cascade="all, delete-orphan",
                cascade_backrefs=False,
            ),
            cascade_backrefs=False,
        )

    def __repr__(self):
        return (
            self.__class__.__name__
            + " ("
            + str(self.keys)
            + " = "
            + str(self.values)
            + ")"
        )



from builtins import object

from sqlalchemy.sql import select

from fonduer.parser.models.context import Context


class TemporaryContext(object):
    """
    A context which does not incur the overhead of a proper ORM-based Context
    object. The TemporaryContext class is specifically for the candidate
    extraction process, during which a MentionSpace object will generate many
    TemporaryContexts, which will then be filtered by Matchers prior to
    materialization of Mentions and constituent Context objects.

    Every Context object has a corresponding TemporaryContext object from which
    it inherits.

    A TemporaryContext must have specified equality / set membership semantics,
    a stable_id for checking uniqueness against the database, and a promote()
    method which returns a corresponding Context object.
    """

    def __init__(self):
        self.id = None

    def _load_id_or_insert(self, session):
        """Load the id of the temporary context if it exists or return insert args.

        As a side effect, this also inserts the Context object for the stableid.

        :return: The record of the temporary context to insert.
        :rtype: dict
        """
        if self.id is None:
            stable_id = self.get_stable_id()
            # Check if exists
            id = session.execute(
                select([Context.id]).where(Context.stable_id == stable_id)
            ).first()

            # If not, insert
            if id is None:
                self.id = session.execute(
                    Context.__table__.insert(),
                    {"type": self._get_table().__tablename__, "stable_id": stable_id},
                ).inserted_primary_key[0]
                insert_args = self._get_insert_args()
                insert_args["id"] = self.id
                return insert_args
            else:
                self.id = id[0]

    def __eq__(self, other):
        raise NotImplementedError()

    def __ne__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()

    def _get_polymorphic_identity(self):
        raise NotImplementedError()

    def get_stable_id(self):
        raise NotImplementedError()

    def _get_table(self):
        raise NotImplementedError()

    def _get_insert_args(self):
        raise NotImplementedError()



from builtins import range, str

from treedlib import (
    Children,
    Compile,
    Indicator,
    LeftNgrams,
    LeftSiblings,
    Mention,
    Ngrams,
    Parents,
    RightNgrams,
    RightSiblings,
    compile_relation_feature_generator,
)

from fonduer.candidates.models.span import TemporarySpan
from fonduer.features.feature_libs.tree_structs import corenlp_to_xmltree
from fonduer.utils.config import get_config
from fonduer.utils.data_model_utils import get_left_ngrams, get_right_ngrams
from fonduer.utils.utils import get_as_dict, tokens_to_ngrams

DEF_VALUE = 1

unary_ddlib_feats = {}
unary_word_feats = {}
unary_tdl_feats = {}
binary_tdl_feats = {}
settings = get_config()


def get_content_feats(candidates):
    candidates = candidates if isinstance(candidates, list) else [candidates]
    for candidate in candidates:
        args = tuple([m.span for m in candidate.get_mentions()])
        if not (isinstance(args[0], TemporarySpan)):
            raise ValueError(
                "Accepts Span-type arguments, %s-type found." % type(candidate)
            )

        # Unary candidates
        if len(args) == 1:
            span = args[0]
            if span.sentence.is_lingual():
                get_tdl_feats = compile_entity_feature_generator()
                sent = get_as_dict(span.sentence)
                xmltree = corenlp_to_xmltree(sent)
                sidxs = list(
                    range(span.get_word_start_index(), span.get_word_end_index() + 1)
                )
                if len(sidxs) > 0:
                    # Add DDLIB entity features
                    for f in get_ddlib_feats(span, sent, sidxs):
                        yield candidate.id, "DDL_" + f, DEF_VALUE
                    # Add TreeDLib entity features
                    if span.stable_id not in unary_tdl_feats:
                        unary_tdl_feats[span.stable_id] = set()
                        for f in get_tdl_feats(xmltree.root, sidxs):
                            unary_tdl_feats[span.stable_id].add(f)
                    for f in unary_tdl_feats[span.stable_id]:
                        yield candidate.id, "TDL_" + f, DEF_VALUE
            else:
                for f in get_word_feats(span):
                    yield candidate.id, "BASIC_" + f, DEF_VALUE

        # Binary candidates
        elif len(args) == 2:
            span1, span2 = args
            if span1.sentence.is_lingual() and span2.sentence.is_lingual():
                get_tdl_feats = compile_relation_feature_generator()
                sent1 = get_as_dict(span1.sentence)
                sent2 = get_as_dict(span2.sentence)
                xmltree = corenlp_to_xmltree(get_as_dict(span1.sentence))
                s1_idxs = list(
                    range(span1.get_word_start_index(), span1.get_word_end_index() + 1)
                )
                s2_idxs = list(
                    range(span2.get_word_start_index(), span2.get_word_end_index() + 1)
                )
                if len(s1_idxs) > 0 and len(s2_idxs) > 0:

                    # Add DDLIB entity features for relation
                    for f in get_ddlib_feats(span1, sent1, s1_idxs):
                        yield candidate.id, "DDL_e1_" + f, DEF_VALUE

                    for f in get_ddlib_feats(span2, sent2, s2_idxs):
                        yield candidate.id, "DDL_e2_" + f, DEF_VALUE

                    # Add TreeDLib relation features
                    if candidate.id not in binary_tdl_feats:
                        binary_tdl_feats[candidate.id] = set()
                        for f in get_tdl_feats(xmltree.root, s1_idxs, s2_idxs):
                            binary_tdl_feats[candidate.id].add(f)
                    for f in binary_tdl_feats[candidate.id]:
                        yield candidate.id, "TDL_" + f, DEF_VALUE
            else:
                for f in get_word_feats(span1):
                    yield candidate.id, "BASIC_e1_" + f, DEF_VALUE

                for f in get_word_feats(span2):
                    yield candidate.id, "BASIC_e2_" + f, DEF_VALUE

        else:
            raise NotImplementedError(
                "Only handles unary and binary candidates currently"
            )


def compile_entity_feature_generator():
    """
    Given optional arguments, returns a generator function which accepts an xml
    root and a list of indexes for a mention, and will generate relation
    features for this entity.
    """

    BASIC_ATTRIBS_REL = ["lemma", "dep_label"]

    m = Mention(0)

    # Basic relation feature templates
    temps = [
        [Indicator(m, a) for a in BASIC_ATTRIBS_REL],
        Indicator(m, "dep_label,lemma"),
        # The *first element on the* path to the root: ngram lemmas along it
        Ngrams(Parents(m, 3), "lemma", (1, 3)),
        Ngrams(Children(m), "lemma", (1, 3)),
        # The siblings of the mention
        [LeftNgrams(LeftSiblings(m), a) for a in BASIC_ATTRIBS_REL],
        [RightNgrams(RightSiblings(m), a) for a in BASIC_ATTRIBS_REL],
    ]

    # return generator function
    return Compile(temps).apply_mention


def get_ddlib_feats(span, context, idxs):
    """
    Minimalist port of generic mention features from ddlib
    """

    if span.stable_id not in unary_ddlib_feats:
        unary_ddlib_feats[span.stable_id] = set()

        for seq_feat in _get_seq_features(context, idxs):
            unary_ddlib_feats[span.stable_id].add(seq_feat)

        for window_feat in _get_window_features(context, idxs):
            unary_ddlib_feats[span.stable_id].add(window_feat)

    for f in unary_ddlib_feats[span.stable_id]:
        yield f


def _get_seq_features(context, idxs):
    yield "WORD_SEQ_[" + " ".join(context["words"][i] for i in idxs) + "]"
    yield "LEMMA_SEQ_[" + " ".join(context["lemmas"][i] for i in idxs) + "]"
    yield "POS_SEQ_[" + " ".join(context["pos_tags"][i] for i in idxs) + "]"
    yield "DEP_SEQ_[" + " ".join(context["dep_labels"][i] for i in idxs) + "]"


def _get_window_features(
    context,
    idxs,
    window=settings["featurization"]["content"]["window_feature"]["size"],
    combinations=settings["featurization"]["content"]["window_feature"]["combinations"],
    isolated=settings["featurization"]["content"]["window_feature"]["isolated"],
):
    left_lemmas = []
    left_pos_tags = []
    right_lemmas = []
    right_pos_tags = []
    try:
        for i in range(1, window + 1):
            lemma = context["lemmas"][idxs[0] - i]
            try:
                float(lemma)
                lemma = "_NUMBER"
            except ValueError:
                pass
            left_lemmas.append(lemma)
            left_pos_tags.append(context["pos_tags"][idxs[0] - i])
    except IndexError:
        pass
    left_lemmas.reverse()
    left_pos_tags.reverse()
    try:
        for i in range(1, window + 1):
            lemma = context["lemmas"][idxs[-1] + i]
            try:
                float(lemma)
                lemma = "_NUMBER"
            except ValueError:
                pass
            right_lemmas.append(lemma)
            right_pos_tags.append(context["pos_tags"][idxs[-1] + i])
    except IndexError:
        pass
    if isolated:
        for i in range(len(left_lemmas)):
            yield "W_LEFT_" + str(i + 1) + "_[" + " ".join(left_lemmas[-i - 1 :]) + "]"
            yield "W_LEFT_POS_" + str(i + 1) + "_[" + " ".join(
                left_pos_tags[-i - 1 :]
            ) + "]"
        for i in range(len(right_lemmas)):
            yield "W_RIGHT_" + str(i + 1) + "_[" + " ".join(right_lemmas[: i + 1]) + "]"
            yield "W_RIGHT_POS_" + str(i + 1) + "_[" + " ".join(
                right_pos_tags[: i + 1]
            ) + "]"
    if combinations:
        for i in range(len(left_lemmas)):
            curr_left_lemmas = " ".join(left_lemmas[-i - 1 :])
            try:
                curr_left_pos_tags = " ".join(left_pos_tags[-i - 1 :])
            except TypeError:
                new_pos_tags = []
                for pos in left_pos_tags[-i - 1 :]:
                    to_add = pos
                    if not to_add:
                        to_add = "None"
                    new_pos_tags.append(to_add)
                curr_left_pos_tags = " ".join(new_pos_tags)
            for j in range(len(right_lemmas)):
                curr_right_lemmas = " ".join(right_lemmas[: j + 1])
                try:
                    curr_right_pos_tags = " ".join(right_pos_tags[: j + 1])
                except TypeError:
                    new_pos_tags = []
                    for pos in right_pos_tags[: j + 1]:
                        to_add = pos
                        if not to_add:
                            to_add = "None"
                        new_pos_tags.append(to_add)
                    curr_right_pos_tags = " ".join(new_pos_tags)
                yield "W_LEMMA_L_" + str(i + 1) + "_R_" + str(
                    j + 1
                ) + "_[" + curr_left_lemmas + "]_[" + curr_right_lemmas + "]"
                yield "W_POS_L_" + str(i + 1) + "_R_" + str(
                    j + 1
                ) + "_[" + curr_left_pos_tags + "]_[" + curr_right_pos_tags + "]"


def get_word_feats(span):
    attrib = "words"

    if span.stable_id not in unary_word_feats:
        unary_word_feats[span.stable_id] = set()

        for ngram in tokens_to_ngrams(span.get_attrib_tokens(attrib), n_min=1, n_max=2):
            feature = "CONTAINS_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

        for ngram in get_left_ngrams(
            span,
            window=settings["featurization"]["content"]["word_feature"]["window"],
            n_max=2,
            attrib=attrib,
        ):
            feature = "LEFT_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

        for ngram in get_right_ngrams(
            span,
            window=settings["featurization"]["content"]["word_feature"]["window"],
            n_max=2,
            attrib=attrib,
        ):
            feature = "RIGHT_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

    for f in unary_word_feats[span.stable_id]:
        yield f



import logging

from fonduer.candidates.models import Candidate
from fonduer.supervision.models import GoldLabelKey, Label, LabelKey
from fonduer.utils.udf import UDF, UDFRunner
from fonduer.utils.utils_udf import (
    ALL_SPLITS,
    add_keys,
    batch_upsert_records,
    get_cands_list_from_split,
    get_docs_from_split,
    get_mapping,
    get_sparse_matrix,
    get_sparse_matrix_keys,
)

logger = logging.getLogger(__name__)


def get_gold_labels(session, cand_lists, annotator_name="gold"):
    """Get the sparse matrix for the specified annotator.

    :param session: The database session.
    :param cand_lists: The candidates to get gold labels for.
    :type cand_lists: List of list of candidates.
    :param annotator: A specific annotator key to get labels for. Default
        "gold".
    :type annotator: str
    """
    return get_sparse_matrix(session, GoldLabelKey, cand_lists, key=annotator_name)


class Labeler(UDFRunner):
    """An operator to add Label Annotations to Candidates.

    :param session: The database session to use.
    :param candidate_classes: A list of candidate_subclasses to label.
    :type candidate_classes: list
    :param parallelism: The number of processes to use in parallel. Default 1.
    :type parallelism: int
    """

    def __init__(self, session, candidate_classes, parallelism=1):
        """Initialize the Labeler."""
        super(Labeler, self).__init__(
            session,
            LabelerUDF,
            parallelism=parallelism,
            candidate_classes=candidate_classes,
        )
        self.candidate_classes = candidate_classes
        self.lfs = []

    def update(self, docs=None, split=0, lfs=None, parallelism=None, progress_bar=True):
        """Update the labels of the specified candidates based on the provided LFs.

        :param docs: If provided, apply the updated LFs to all the candidates
            in these documents.
        :param split: If docs is None, apply the updated LFs to the candidates
            in this particular split.
        :param lfs: A list of lists of labeling functions to update. Each list
            should correspond with the candidate_classes used to initialize the
            Labeler.
        :param parallelism: How many threads to use for extraction. This will
            override the parallelism value used to initialize the Labeler if
            it is provided.
        :type parallelism: int
        :param progress_bar: Whether or not to display a progress bar. The
            progress bar is measured per document.
        :type progress_bar: bool
        """
        if lfs is None:
            raise ValueError("Please provide a list of lists of labeling functions.")

        if len(lfs) != len(self.candidate_classes):
            raise ValueError("Please provide LFs for each candidate class.")

        self.apply(
            docs=docs,
            split=split,
            lfs=lfs,
            train=True,
            clear=False,
            parallelism=parallelism,
            progress_bar=progress_bar,
        )

    def apply(
        self,
        docs=None,
        split=0,
        train=False,
        lfs=None,
        clear=True,
        parallelism=None,
        progress_bar=True,
    ):
        """Apply the labels of the specified candidates based on the provided LFs.

        :param docs: If provided, apply the LFs to all the candidates in these
            documents.
        :param split: If docs is None, apply the LFs to the candidates in this
            particular split.
        :type split: int
        :param train: Whether or not to update the global key set of labels and
            the labels of candidates.
        :type train: bool
        :param lfs: A list of lists of labeling functions to apply. Each list
            should correspond with the candidate_classes used to initialize the
            Labeler.
        :type lfs: list of lists
        :param clear: Whether or not to clear the labels table before applying
            these LFs.
        :type clear: bool
        :param parallelism: How many threads to use for extraction. This will
            override the parallelism value used to initialize the Labeler if
            it is provided.
        :type parallelism: int
        :param progress_bar: Whether or not to display a progress bar. The
            progress bar is measured per document.
        :type progress_bar: bool

        :raises ValueError: If labeling functions are not provided for each
            candidate class.
        """
        if lfs is None:
            raise ValueError("Please provide a list of labeling functions.")

        if len(lfs) != len(self.candidate_classes):
            raise ValueError("Please provide LFs for each candidate class.")

        self.lfs = lfs
        if docs:
            # Call apply on the specified docs for all splits
            split = ALL_SPLITS
            super(Labeler, self).apply(
                docs,
                split=split,
                train=train,
                lfs=self.lfs,
                clear=clear,
                parallelism=parallelism,
                progress_bar=progress_bar,
            )
            # Needed to sync the bulk operations
            self.session.commit()
        else:
            # Only grab the docs containing candidates from the given split.
            split_docs = get_docs_from_split(
                self.session, self.candidate_classes, split
            )
            super(Labeler, self).apply(
                split_docs,
                split=split,
                train=train,
                lfs=self.lfs,
                clear=clear,
                parallelism=parallelism,
                progress_bar=progress_bar,
            )
            # Needed to sync the bulk operations
            self.session.commit()

    def get_keys(self):
        """Return a list of keys for the Labels.

        :return: List of LabelKeys.
        :rtype: list
        """
        return list(get_sparse_matrix_keys(self.session, LabelKey))

    def drop_keys(self, keys):
        """Drop the specified keys from LabelKeys.

        :param keys: A list of labeling functions to delete.
        :type keys: list, tuple
        """
        # Make sure keys is iterable
        keys = keys if isinstance(keys, (list, tuple)) else [keys]

        # Remove the specified keys
        for key in keys:
            try:  # Assume key is an LF
                self.session.query(LabelKey).filter(
                    LabelKey.name == key.__name__
                ).delete()
            except AttributeError:
                self.session.query(LabelKey).filter(LabelKey.name == key).delete()

    def clear(self, train, split, lfs=None):
        """Delete Labels of each class from the database.

        :param train: Whether or not to clear the LabelKeys.
        :type train: bool
        :param split: Which split of candidates to clear labels from.
        :type split: int
        :param lfs: This parameter is ignored.
        """
        # Clear Labels for the candidates in the split passed in.
        logger.info("Clearing Labels (split {})".format(split))

        sub_query = (
            self.session.query(Candidate.id).filter(Candidate.split == split).subquery()
        )
        query = self.session.query(Label).filter(Label.candidate_id.in_(sub_query))
        query.delete(synchronize_session="fetch")

        # Delete all old annotation keys
        if train:
            logger.debug("Clearing all LabelKey...")
            query = self.session.query(LabelKey)
            query.delete(synchronize_session="fetch")

    def clear_all(self):
        """Delete all Labels."""
        logger.info("Clearing ALL Labels and LabelKeys.")
        self.session.query(Label).delete()
        self.session.query(LabelKey).delete()

    def get_gold_labels(self, cand_lists, annotator=None):
        """Load sparse matrix of GoldLabels for each candidate_class.

        :param cand_lists: The candidates to get gold labels for.
        :type cand_lists: List of list of candidates.
        :param annotator: A specific annotator key to get labels for. Default
            None.
        :type annotator: str
        :return: An MxN sparse matrix where M are the candidates and N is the
            annotators. If annotator is provided, return an Mx1 matrix.
        :rtype: csr_matrix
        """
        return get_sparse_matrix(self.session, GoldLabelKey, cand_lists, key=annotator)

    def get_label_matrices(self, cand_lists):
        """Load sparse matrix of Labels for each candidate_class.

        :param cand_lists: The candidates to get labels for.
        :type cand_lists: List of list of candidates.
        :return: An MxN sparse matrix where M are the candidates and N is the
            labeling functions.
        :rtype: csr_matrix
        """
        return get_sparse_matrix(self.session, LabelKey, cand_lists)


class LabelerUDF(UDF):
    """UDF for performing candidate extraction."""

    def __init__(self, candidate_classes, **kwargs):
        """Initialize the LabelerUDF."""
        self.candidate_classes = (
            candidate_classes
            if isinstance(candidate_classes, (list, tuple))
            else [candidate_classes]
        )
        super(LabelerUDF, self).__init__(**kwargs)

    def _f_gen(self, c):
        """Convert lfs into a generator of id, name, and labels.

        In particular, catch verbose values and convert to integer ones.
        """
        lf_idx = self.candidate_classes.index(c.__class__)
        labels = lambda c: [(c.id, lf.__name__, lf(c)) for lf in self.lfs[lf_idx]]
        for cid, lf_key, label in labels(c):
            # Note: We assume if the LF output is an int, it is already
            # mapped correctly
            if isinstance(label, int):
                yield cid, lf_key, label
            # None is a protected LF output value corresponding to 0,
            # representing LF abstaining
            elif label is None:
                yield cid, lf_key, 0
            elif label in c.values:
                if c.cardinality > 2:
                    yield cid, lf_key, c.values.index(label) + 1
                # Note: Would be nice to not special-case here, but for
                # consistency we leave binary LF range as {-1,0,1}
                else:
                    val = 1 if c.values.index(label) == 0 else -1
                    yield cid, lf_key, val
            else:
                raise ValueError(
                    "Can't parse label value {} for candidate values {}".format(
                        label, c.values
                    )
                )

    def apply(self, doc, split, train, lfs, **kwargs):
        """Extract candidates from the given Context.

        :param doc: A document to process.
        :param split: Which split to use.
        :param train: Whether or not to insert new LabelKeys.
        :param lfs: The list of functions to use to generate labels.
        """
        logger.debug("Document: {}".format(doc))

        if lfs is None:
            raise ValueError("Must provide lfs kwarg.")

        self.lfs = lfs

        # Get all the candidates in this doc that will be featurized
        cands_list = get_cands_list_from_split(
            self.session, self.candidate_classes, doc, split
        )

        label_keys = set()
        for cands in cands_list:
            records = list(
                get_mapping(self.session, Label, cands, self._f_gen, label_keys)
            )
            batch_upsert_records(self.session, Label, records)

        # Insert all Label Keys
        if train:
            add_keys(self.session, LabelKey, label_keys)

        # This return + yield makes a completely empty generator
        return
        yield



"""
A sparse linear module.
"""

import math

import torch
import torch.nn as nn


class SparseLinear(nn.Module):
    """A sparse linear layer.

    :param num_features: Size of features.
    :type num_features: int
    :param num_classes: Number of classes.
    :type num_classes: int
    :param bias: Use bias term or not.
    :type bias: bool
    :param padding_idx: padding index.
    :type padding_idx: int
    """

    def __init__(self, num_features, num_classes, bias=False, padding_idx=0):

        super(SparseLinear, self).__init__()

        self.num_features = num_features
        self.num_classes = num_classes
        self.padding_idx = padding_idx

        self.weight = nn.Embedding(
            self.num_features, self.num_classes, padding_idx=self.padding_idx
        )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(self.num_classes))
        else:
            self.bias = None

        self.reset_parameters()

    def reset_parameters(self):
        """Reinitiate the weight parameters.
        """

        stdv = 1.0 / math.sqrt(self.num_features)
        self.weight.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)
        if self.padding_idx is not None:
            self.weight.weight.data[self.padding_idx].fill_(0)

    def forward(self, x, w):
        """Forward function.

        :param x: Feature indices.
        :type x: torch.Tensor of shape (batch_size * length)
        :param w: Feature weights.
        :type w: torch.Tensor of shape (batch_size * length)
        :return: Output of linear layer.
        :rtype: torch.Tensor of shape (batch_size, num_classes)
        """

        if self.bias is None:
            return (w.unsqueeze(2) * self.weight(x)).sum(dim=1)
        else:
            return (w.unsqueeze(2) * self.weight(x)).sum(dim=1) + self.bias



import logging
from builtins import object
from urllib.parse import urlparse

import psycopg2
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

logger = logging.getLogger(__name__)


# Defines procedure for setting up a sessionmaker
def new_sessionmaker():
    # Turning on autocommit for Postgres, see
    # http://oddbird.net/2014/06/14/sqlalchemy-postgres-autocommit/
    # Otherwise any e.g. query starts a transaction, locking tables... very
    # bad for e.g. multiple notebooks open, multiple processes, etc.
    if Meta.postgres and Meta.ready:
        engine = create_engine(
            Meta.conn_string,
            client_encoding="utf8",
            use_batch_mode=True,
            isolation_level="AUTOCOMMIT",
        )
    else:
        raise ValueError(
            "Meta variables have not been initialized with "
            "a valid postgres connection string."
        )
    # New sessionmaker
    return sessionmaker(bind=engine)


def _validate_conn_string(conn_string):
    """Check that the PostgreSQL connection string is valid."""
    logger.debug("Validating {} as a connection string...".format(conn_string))
    url = urlparse(conn_string)
    Meta.conn_string = conn_string
    Meta.DBNAME = url.path[1:]
    Meta.DBUSER = url.username
    Meta.DBPWD = url.password
    Meta.DBHOST = url.hostname
    Meta.DBPORT = url.port
    Meta.postgres = url.scheme.startswith("postgres")
    # Actually try to connect to see if the connection string is valid
    try:
        con = psycopg2.connect(
            database=Meta.DBNAME,
            user=Meta.DBUSER,
            password=Meta.DBPWD,
            host=Meta.DBHOST,
        )
        con.close()
    except psycopg2.OperationalError as e:
        raise ValueError(
            "{} is an invalid connection string. Use the form {}".format(
                conn_string, "postgres://<user>:<pw>@<host>:<port>/<database_name>"
            )
        )
    return True


class Meta(object):
    """Singleton-like metadata class for all global variables.

    Adapted from the Unique Design Pattern:
    https://stackoverflow.com/questions/1318406/why-is-the-borg-pattern-better-than-the-singleton-pattern-in-python
    """

    # Static class variables
    conn_string = None
    DBNAME = None
    DBUSER = None
    DBHOST = None
    DBPORT = None
    DBPWD = None
    Session = None
    engine = None
    Base = declarative_base(name="Base", cls=object)
    postgres = False
    ready = False

    @classmethod
    def init(cls, conn_string=None):
        """Return the unique Meta class."""
        if conn_string:
            Meta.ready = _validate_conn_string(conn_string)
            # We initialize the engine within the models module because models'
            # schema can depend on which data types are supported by the engine
            Meta.Session = new_sessionmaker()
            Meta.engine = Meta.Session.kw["bind"]
            logger.info(
                "Connecting user:{} to {}:{}/{}".format(
                    Meta.DBUSER, Meta.DBHOST, Meta.DBPORT, Meta.DBNAME
                )
            )
            Meta._init_db()

        return cls

    @classmethod
    def _init_db(cls):
        """ Initialize the storage schema.

        This call must be performed after all classes that extend
        Base are declared to ensure the storage schema is initialized.
        """
        # This list of import defines which SQLAlchemy classes will be
        # initialized when Meta.init() is called. If a sqlalchemy class is not
        # imported before the call to create_all(), it will not be created.
        import fonduer.candidates.models  # noqa
        import fonduer.features.models  # noqa
        import fonduer.learning.models  # noqa
        import fonduer.parser.models  # noqa
        import fonduer.supervision.models  # noqa
        import fonduer.utils.models  # noqa

        logger.info("Initializing the storage schema")
        Meta.Base.metadata.create_all(Meta.engine)



from sqlalchemy import Column, Integer, String

from fonduer.meta import Meta

# Grab pointer to global metadata
_meta = Meta.init()


class Context(_meta.Base):
    """A piece of content from which Candidates are composed.

    This serves as the base class of the Fonduer document model.
    """

    __tablename__ = "context"

    #: The unique id of the ``Context``.
    id = Column(Integer, primary_key=True)

    #: The type of the ``Context`` represented as a string (e.g. "sentence",
    #: "paragraph", "figure").
    type = Column(String, nullable=False)

    #: A stable representation of the ``Context`` that will not change between
    #: runs.
    stable_id = Column(String, unique=True, nullable=False)

    __mapper_args__ = {"polymorphic_identity": "context", "polymorphic_on": type}



from sqlalchemy import Column, Float
from sqlalchemy.dialects import postgresql

from fonduer.meta import Meta
from fonduer.utils.models.annotation import AnnotationKeyMixin, AnnotationMixin

_meta = Meta.init()


class FeatureKey(AnnotationKeyMixin, _meta.Base):
    """A feature's key that identifies the definition of the Feature."""

    pass


class Feature(AnnotationMixin, _meta.Base):
    """An element of a representation of a Candidate in a feature space.

    A Feature's annotation key identifies the definition of the Feature, e.g.,
    a function that implements it or the library name and feature name in an
    automatic featurization library.
    """

    #: A list of floating point values for each Key.
    values = Column(postgresql.ARRAY(Float), nullable=False)



import logging
from builtins import range
from itertools import product

from sqlalchemy.sql import select

from fonduer.candidates.models import Candidate
from fonduer.utils.udf import UDF, UDFRunner

logger = logging.getLogger(__name__)


class CandidateExtractor(UDFRunner):
    """An operator to extract Candidate objects from a Context.

    :Example:

        Assuming we have already defined a Part and Temp ``Mention`` subclass,
        and a throttler called templ_throttler, we can create a candidate
        extractor as follows::

            PartTemp = candidate_subclass("PartTemp", [Part, Temp])
            candidate_extractor = CandidateExtractor(
                session, [PartTemp], throttlers=[temp_throttler]
            )

    :param session: An initialized database session.
    :param candidate_classes: The types of relation to extract, defined using
        :func: fonduer.candidates.candidate_subclass.
    :type candidate_classes: list of candidate subclasses.
    :param throttlers: optional functions for filtering out candidates
        which returns a Boolean expressing whether or not the candidate should
        be instantiated.
    :type throttlers: list of throttlers.
    :param self_relations: Boolean indicating whether to extract Candidates
        that relate the same context. Only applies to binary relations.
    :type self_relations: bool
    :param nested_relations: Boolean indicating whether to extract Candidates
        that relate one Context with another that contains it. Only applies to
        binary relations.
    :type nested_relations: bool
    :param symmetric_relations: Boolean indicating whether to extract symmetric
        Candidates, i.e., rel(A,B) and rel(B,A), where A and B are Contexts.
        Only applies to binary relations.
    :type symmetric_relations: bool
    :param parallelism: The number of processes to use in parallel for calls
        to apply().
    :type parallelism: int
    :raises ValueError: If throttlers are provided, but a throtters are not the
        same length as candidate classes.
    """

    def __init__(
        self,
        session,
        candidate_classes,
        throttlers=None,
        self_relations=False,
        nested_relations=False,
        symmetric_relations=True,
        parallelism=1,
    ):
        """ Set throttlers match candidate_classes if not provide. """
        if throttlers is None:
            throttlers = [None] * len(candidate_classes)

        """Initialize the CandidateExtractor."""
        super(CandidateExtractor, self).__init__(
            session,
            CandidateExtractorUDF,
            parallelism=parallelism,
            candidate_classes=candidate_classes,
            throttlers=throttlers,
            self_relations=self_relations,
            nested_relations=nested_relations,
            symmetric_relations=symmetric_relations,
        )
        # Check that arity is sensible
        if len(candidate_classes) != len(throttlers):
            raise ValueError(
                "Provided different number of throttlers and candidate classes."
            )

        self.candidate_classes = candidate_classes

    def apply(self, docs, split=0, clear=True, parallelism=None, progress_bar=True):
        """Run the CandidateExtractor.

        :Example: To extract candidates from a set of training documents using
            4 cores::

                candidate_extractor.apply(train_docs, split=0, parallelism=4)

        :param docs: Set of documents to extract from.
        :param split: Which split to assign the extracted Candidates to.
        :type split: int
        :param clear: Whether or not to clear the existing Candidates
            beforehand.
        :type clear: bool
        :param parallelism: How many threads to use for extraction. This will
            override the parallelism value used to initialize the
            CandidateExtractor if it is provided.
        :type parallelism: int
        :param progress_bar: Whether or not to display a progress bar. The
            progress bar is measured per document.
        :type progress_bar: bool
        """
        super(CandidateExtractor, self).apply(
            docs,
            split=split,
            clear=clear,
            parallelism=parallelism,
            progress_bar=progress_bar,
        )

    def clear(self, split):
        """Delete Candidates of each class initialized with the
        CandidateExtractor from given split the database.

        :param split: Which split to clear.
        :type split: int
        """
        for candidate_class in self.candidate_classes:
            logger.info(
                "Clearing table {} (split {})".format(
                    candidate_class.__tablename__, split
                )
            )
            self.session.query(Candidate).filter(
                Candidate.type == candidate_class.__tablename__
            ).filter(Candidate.split == split).delete()

    def clear_all(self, split):
        """Delete ALL Candidates from given split the database.

        :param split: Which split to clear.
        :type split: int
        """
        logger.info("Clearing ALL Candidates.")
        self.session.query(Candidate).filter(Candidate.split == split).delete()

    def get_candidates(self, docs=None, split=0):
        """Return a list of lists of the candidates associated with this extractor.

        Each list of the return will contain the candidates for one of the
        candidate classes associated with the CandidateExtractor.

        :param docs: If provided, return candidates from these documents from
            all splits.
        :type docs: list, tuple of ``Documents``.
        :param split: If docs is None, then return all the candidates from this
            split.
        :type split: int
        :return: Candidates for each candidate_class.
        :rtype: List of lists of ``Candidates``.
        """
        result = []
        if docs:
            docs = docs if isinstance(docs, (list, tuple)) else [docs]
            # Get cands from all splits
            for candidate_class in self.candidate_classes:
                cands = (
                    self.session.query(candidate_class)
                    .filter(candidate_class.document_id.in_([doc.id for doc in docs]))
                    .order_by(candidate_class.id)
                    .all()
                )
                result.append(cands)
        else:
            for candidate_class in self.candidate_classes:
                # Filter by candidate_ids in a particular split
                sub_query = (
                    self.session.query(Candidate.id)
                    .filter(Candidate.split == split)
                    .subquery()
                )
                cands = (
                    self.session.query(candidate_class)
                    .filter(candidate_class.id.in_(sub_query))
                    .order_by(candidate_class.id)
                    .all()
                )
                result.append(cands)
        return result


class CandidateExtractorUDF(UDF):
    """UDF for performing candidate extraction."""

    def __init__(
        self,
        candidate_classes,
        throttlers,
        self_relations,
        nested_relations,
        symmetric_relations,
        **kwargs
    ):
        """Initialize the CandidateExtractorUDF."""
        self.candidate_classes = (
            candidate_classes
            if isinstance(candidate_classes, (list, tuple))
            else [candidate_classes]
        )
        self.throttlers = (
            throttlers if isinstance(throttlers, (list, tuple)) else [throttlers]
        )
        self.nested_relations = nested_relations
        self.self_relations = self_relations
        self.symmetric_relations = symmetric_relations
        self.arities = [len(cclass.__argnames__) for cclass in self.candidate_classes]

        super(CandidateExtractorUDF, self).__init__(**kwargs)

    def apply(self, context, clear, split, **kwargs):
        """Extract candidates from the given Context.

        :param context: A document to process.
        :param clear: Whether or not to clear the existing database entries.
        :param split: Which split to use.
        """
        logger.debug("Document: {}".format(context))
        # Iterate over each candidate class
        for i, candidate_class in enumerate(self.candidate_classes):
            logger.debug("  Relation: {}".format(candidate_class.__name__))
            # Generates and persists candidates
            candidate_args = {"split": split}
            candidate_args["document_id"] = context.id
            cands = product(
                *[
                    enumerate(
                        self.session.query(mention)
                        .filter(mention.document_id == context.id)
                        .order_by(mention.id)
                        .all()
                    )
                    for mention in candidate_class.mentions
                ]
            )
            for cand in cands:

                # Apply throttler if one was given.
                # Accepts a tuple of Mention objects
                # (throttler returns whether or not proposed candidate
                # passes throttling condition)
                if self.throttlers[i]:
                    if not self.throttlers[i](
                        tuple(cand[j][1] for j in range(self.arities[i]))
                    ):
                        continue

                # TODO: Make this work for higher-order relations
                if self.arities[i] == 2:
                    ai, a = (cand[0][0], cand[0][1].span)
                    bi, b = (cand[1][0], cand[1][1].span)

                    # Check for self-joins, "nested" joins (joins from span to
                    # its subspan), and flipped duplicate "symmetric" relations
                    if not self.self_relations and a == b:
                        logger.debug("Skipping self-joined candidate {}".format(cand))
                        continue
                    if not self.nested_relations and (a in b or b in a):
                        logger.debug("Skipping nested candidate {}".format(cand))
                        continue
                    if not self.symmetric_relations and ai > bi:
                        logger.debug("Skipping symmetric candidate {}".format(cand))
                        continue

                # Assemble candidate arguments
                for j, arg_name in enumerate(candidate_class.__argnames__):
                    candidate_args[arg_name + "_id"] = cand[j][1].id

                # Checking for existence
                if not clear:
                    q = select([candidate_class.id])
                    for key, value in list(candidate_args.items()):
                        q = q.where(getattr(candidate_class, key) == value)
                    candidate_id = self.session.execute(q).first()
                    if candidate_id is not None:
                        continue

                # Add Candidate to session
                yield candidate_class(**candidate_args)



import numpy as np

from fonduer.learning.utils import save_marginals


class Classifier(object):
    """An abstract class for a probabilistic classifier.

    :param name: Name of the model
    :type name: str
    """

    def __init__(self, name=None):
        self.name = name or self.__class__.__name__
        self.cardinality = 2

    def marginals(self, X, batch_size=None):
        """Calculate the predicted marginal probabilities for the Candidates X.

        :param X: Input data.
        :param batch_size: Batch size.
        :type batch_size: int
        """

        raise NotImplementedError()

    def save_marginals(self, session, X, training=False):
        """Save the predicted marginal probabilities for the Candidates X.

        :param session: The database session to use.
        :param X: Input data.
        :param training: If True, these are training marginals / labels;
            else they are saved as end model predictions.
        :type training: bool
        """

        save_marginals(session, X, self.marginals(X), training=training)

    def predictions(self, X, b=0.5, batch_size=None):
        """Return numpy array of elements in {-1,0,1}
        based on predicted marginal probabilities.

        :param X: Input data.
        :param b: Prediction threshold.
        :type b: float
        :param batch_size: Batch size.
        :type batch_size: int
        """
        if self._check_input(X):
            X = self._preprocess_data(X)

        if self.cardinality > 2:
            return self.marginals(X, batch_size=batch_size).argmax(axis=1) + 1
        else:
            return np.array(
                [
                    1 if p > b else -1 if p < b else 0
                    for p in self.marginals(X, batch_size=batch_size)
                ]
            )

    def score(
        self, X_test, Y_test, b=0.5, set_unlabeled_as_neg=True, beta=1, batch_size=None
    ):
        """
        Returns the summary scores:
            * For binary: precision, recall, F-beta score
            * For categorical: accuracy

        :param X_test: The input test candidates.
        :type X_test: pair with candidates and corresponding features
        :param Y_test: The input test labels.
        :type Y_test: list of labels
        :param b: Decision boundary *for binary setting only*.
        :type b: float
        :param set_unlabeled_as_neg: Whether to map 0 labels -> -1,
            *binary setting.*
        :type set_unlabeled_as_neg: bool
        :param beta: For F-beta score; by default beta = 1 => F-1 score.
        :type beta: int
        :param batch_size: Batch size.
        :type batch_size: int
        """

        if self._check_input(X_test):
            X_test, Y_test = self._preprocess_data(X_test, Y_test)

        predictions = self.predictions(X_test, b=b, batch_size=batch_size)

        # Convert Y_test to dense numpy array
        try:
            Y_test = np.array(Y_test.todense()).reshape(-1)
        except Exception:
            Y_test = np.array(Y_test)

        # Compute accuracy for categorical, or P/R/F1 for binary settings
        if self.cardinality > 2:
            # Compute and return accuracy
            correct = np.where([predictions == Y_test])[0].shape[0]
            return correct / float(Y_test.shape[0])
        else:
            # Either remap or filter out unlabeled (0-valued) test labels
            if set_unlabeled_as_neg:
                Y_test[Y_test == 0] = -1
            else:
                predictions = predictions[Y_test != 0]
                Y_test = Y_test[Y_test != 0]

            # Compute and return precision, recall, and F1 score
            tp = (0.5 * (predictions * Y_test + 1))[predictions == 1].sum()
            pred_pos = predictions[predictions == 1].sum()
            p = tp / float(pred_pos) if pred_pos > 0 else 0.0
            pos = Y_test[Y_test == 1].sum()
            r = tp / float(pos) if pos > 0 else 0.0

            # Compute general F-beta score
            if p + r > 0:
                f_beta = (1 + beta ** 2) * ((p * r) / (((beta ** 2) * p) + r))
            else:
                f_beta = 0.0
            return p, r, f_beta

    def _check_input(self, X):
        """Generic data checking function

        :param X: The input data of the model.
        :type X: pair with candidates and corresponding features
        """
        return True

    def _preprocess_data(self, X, Y=None):
        """Generic preprocessing function

        :param X: The input data of the model.
        :type X: pair with candidates and corresponding features
        :param Y: The labels of input data (optional).
        :type Y: list of labels
        """
        if Y is None:
            return X
        return X, Y

    def save(self):
        """Generic model saving function"""
        raise NotImplementedError()

    def load(self):
        """Generic model loading function"""
        raise NotImplementedError()



from fonduer.candidates.models.span import TemporarySpan
from fonduer.utils.data_model_utils import (
    get_visual_aligned_lemmas,
    is_horz_aligned,
    is_vert_aligned,
    is_vert_aligned_center,
    is_vert_aligned_left,
    is_vert_aligned_right,
    same_page,
)

FEAT_PRE = "VIZ_"
DEF_VALUE = 1

unary_vizlib_feats = {}
binary_vizlib_feats = {}


def get_visual_feats(candidates):
    candidates = candidates if isinstance(candidates, list) else [candidates]
    for candidate in candidates:
        args = tuple([m.span for m in candidate.get_mentions()])
        if not (isinstance(args[0], TemporarySpan)):
            raise ValueError(
                "Accepts Span-type arguments, %s-type found." % type(candidate)
            )

        # Unary candidates
        if len(args) == 1:
            span = args[0]
            # Add VisualLib entity features (if applicable)
            if span.sentence.is_visual():
                if span.stable_id not in unary_vizlib_feats:
                    unary_vizlib_feats[span.stable_id] = set()
                    for f, v in vizlib_unary_features(span):
                        unary_vizlib_feats[span.stable_id].add((f, v))

                for f, v in unary_vizlib_feats[span.stable_id]:
                    yield candidate.id, FEAT_PRE + f, v

        # Binary candidates
        elif len(args) == 2:
            span1, span2 = args
            # Add VisualLib entity features (if applicable)
            if span1.sentence.is_visual() or span2.sentence.is_visual():
                for span, pre in [(span1, "e1_"), (span2, "e2_")]:
                    if span.stable_id not in unary_vizlib_feats:
                        unary_vizlib_feats[span.stable_id] = set()
                        for f, v in vizlib_unary_features(span):
                            unary_vizlib_feats[span.stable_id].add((f, v))

                    for f, v in unary_vizlib_feats[span.stable_id]:
                        yield candidate.id, FEAT_PRE + pre + f, v

                if candidate.id not in binary_vizlib_feats:
                    binary_vizlib_feats[candidate.id] = set()
                    for f, v in vizlib_binary_features(span1, span2):
                        binary_vizlib_feats[candidate.id].add((f, v))

                for f, v in binary_vizlib_feats[candidate.id]:
                    yield candidate.id, FEAT_PRE + f, v
        else:
            raise NotImplementedError(
                "Only handles unary and binary candidates currently"
            )


def vizlib_unary_features(span):
    """
    Visual-related features for a single span
    """
    if not span.sentence.is_visual():
        return

    for f in get_visual_aligned_lemmas(span):
        yield "ALIGNED_" + f, DEF_VALUE

    for page in set(span.get_attrib_tokens("page")):
        yield "PAGE_[%d]" % page, DEF_VALUE


def vizlib_binary_features(span1, span2):
    """
    Visual-related features for a pair of spans
    """
    if same_page((span1, span2)):
        yield "SAME_PAGE", DEF_VALUE

        if is_horz_aligned((span1, span2)):
            yield "HORZ_ALIGNED", DEF_VALUE

        if is_vert_aligned((span1, span2)):
            yield "VERT_ALIGNED", DEF_VALUE

        if is_vert_aligned_left((span1, span2)):
            yield "VERT_ALIGNED_LEFT", DEF_VALUE

        if is_vert_aligned_right((span1, span2)):
            yield "VERT_ALIGNED_RIGHT", DEF_VALUE

        if is_vert_aligned_center((span1, span2)):
            yield "VERT_ALIGNED_CENTER", DEF_VALUE



"""For pip."""
from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup

exec(open("src/fonduer/_version.py").read())
setup(
    name="fonduer",
    version=__version__,
    description="Knowledge base construction system for richly formatted data.",
    long_description=open("README.rst").read(),
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    install_requires=[
        "bs4",
        "editdistance",
        "lxml",
        "nltk",
        "numpy>=1.11",
        "pandas",
        "psycopg2-binary",
        "pyyaml",
        "scipy>=0.18",
        "snorkel-metal>=0.1.7",
        "spacy>=2.0.12",
        "sqlalchemy>=1.0.14",
        "torch>=0.4.1",
        "tqdm",
        "treedlib",
        "wand",
    ],
    extras_require={"spacy_ja": ["mecab-python3"]},
    keywords=["fonduer", "knowledge base construction", "richly formatted data"],
    include_package_data=True,
    url="https://github.com/HazyResearch/fonduer",
    classifiers=[  # https://pypi.python.org/pypi?:action=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
    project_urls={
        "Tracker": "https://github.com/HazyResearch/fonduer/issues",
        "Source": "https://github.com/HazyResearch/fonduer",
    },
    python_requires=">=3.6",
    author="Hazy Research",
    author_email="senwu@cs.stanford.edu",
    license="MIT",
)



import logging
import os
from time import time

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from fonduer.learning.classifier import Classifier
from fonduer.learning.utils import (
    LabelBalancer,
    SoftCrossEntropyLoss,
    reshape_marginals,
)


class NoiseAwareModel(Classifier, nn.Module):
    """
    Generic NoiseAwareModel class.

    :param name: Name of the model
    :type name: str
    """

    _gpu = ["gpu", "GPU"]

    def __init__(self, name=None):
        self.logger = logging.getLogger(__name__)
        self.settings = None
        Classifier.__init__(self, name)
        nn.Module.__init__(self)

    def _set_random_seed(self, seed):
        # Set random seed for all numpy operations
        self.rand_state = np.random.RandomState(seed=seed)
        np.random.seed(seed=seed)

        # Set random seed for PyTorch
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    def _build_model(self):
        raise NotImplementedError()

    def _update_settings(self, X):
        pass

    def _preprocess_data(self, X, Y, idxs=None, train=False):
        return X, Y

    def _setup_model_loss(self, lr):
        """
        Setup loss and optimizer for PyTorch model.
        """
        # Setup loss
        if not hasattr(self, "loss"):
            if self.cardinality > 2:
                self.loss = SoftCrossEntropyLoss
            else:
                self.loss = nn.BCEWithLogitsLoss()

        # Setup optimizer
        if not hasattr(self, "optimizer"):
            self.optimizer = optim.Adam(self.parameters(), lr=lr)

    def _check_input(self, X):
        """Checks correctness of input; optional to implement."""
        pass

    def _calc_logits(self, X, batch_size):
        """Calculate the logits of input."""
        raise NotImplementedError()

    def train(
        self,
        X_train,
        Y_train,
        n_epochs=25,
        lr=0.01,
        batch_size=256,
        rebalance=False,
        X_dev=None,
        Y_dev=None,
        print_freq=5,
        dev_ckpt=True,
        dev_ckpt_delay=0.75,
        save_dir="checkpoints",
        seed=1234,
        host_device="CPU",
    ):
        """
        Generic training procedure for PyTorch model

        :param X_train: The training data which is a (list of Candidate objects,
            a sparse matrix of corresponding features) pair.
        :type X_train: pair
        :param Y_train: Array of marginal probabilities for each Candidate.
        :type Y_train: list or numpy.array
        :param n_epochs: Number of training epochs.
        :type n_epochs: int
        :param lr: Learning rate.
        :type lr: float
        :param batch_size: Batch size for learning model.
        :type batch_size: int
        :param rebalance: Bool or fraction of positive examples for training.
                    - if True, defaults to standard 0.5 class balance
                    - if False, no class balancing
        :type rebalance: bool
        :param X_dev: Candidates for evaluation, same format as X_train.
        :param Y_dev: Labels for evaluation, same format as Y_train.
        :param print_freq: number of epochs at which to print status, and if present,
            evaluate the dev set (X_dev, Y_dev).
        :type print_freq: int
        :param dev_ckpt: If True, save a checkpoint whenever highest score
            on (X_dev, Y_dev) reached. Note: currently only evaluates at
            every @print_freq epochs.
        :param dev_ckpt_delay: Start dev checkpointing after this portion
            of n_epochs.
        :type dev_ckpt_delay: float
        :param save_dir: Save dir path for checkpointing.
        :type save_dir: str
        :param seed: Random seed
        :type seed: int
        :param host_device: Host device
        :type host_device: str
        """

        # Set model parameters
        self.settings = {
            "n_epochs": n_epochs,
            "lr": lr,
            "batch_size": batch_size,
            "seed": 1234,
            "host_device": host_device,
        }

        # Set random seed
        self._set_random_seed(self.settings["seed"])

        self._check_input(X_train)
        verbose = print_freq > 0

        # Update cardinality of the model with training marginals
        self.cardinality = Y_train.shape[1] if len(Y_train.shape) > 1 else 2

        # Make sure marginals are in correct default format
        Y_train = reshape_marginals(Y_train)
        # Make sure marginals are in [0,1] (v.s e.g. [-1, 1])
        if self.cardinality > 2 and not np.all(Y_train.sum(axis=1) - 1 < 1e-10):
            raise ValueError("Y_train must be row-stochastic (rows sum to 1).")
        if not np.all(Y_train >= 0):
            raise ValueError("Y_train must have values in [0,1].")

        # Remove unlabeled examples and optionally rebalance training set
        # Note: rebalancing only for binary setting currently
        if self.cardinality == 2:
            # This removes unlabeled examples and optionally rebalances
            train_idxs = LabelBalancer(Y_train).get_train_idxs(
                rebalance, rand_state=self.rand_state
            )
        else:
            # In categorical setting, just remove unlabeled
            diffs = Y_train.max(axis=1) - Y_train.min(axis=1)
            train_idxs = np.where(diffs > 1e-6)[0]

        self._update_settings(X_train)

        _X_train, _Y_train = self._preprocess_data(
            X_train, Y_train, idxs=train_idxs, train=True
        )
        if X_dev is not None:
            _X_dev, _Y_dev = self._preprocess_data(X_dev, Y_dev)

        if self.settings["host_device"] in self._gpu:
            if not torch.cuda.is_available():
                self.settings["host_device"] = "CPU"
                self.logger.info("GPU is not available, switching to CPU...")
            else:
                self.logger.info("Using GPU...")

        self.logger.info("Settings: {}".format(self.settings))

        # Build network
        self._build_model()
        self._setup_model_loss(self.settings["lr"])

        # Set up GPU if necessary
        if self.settings["host_device"] in self._gpu:
            nn.Module.cuda(self)

        # Run mini-batch SGD
        n = len(_X_train)
        if self.settings["batch_size"] > n:
            self.logger.info("Switching batch size to {} for training.".format(n))
        batch_size = min(self.settings["batch_size"], n)

        if verbose:
            st = time()
            self.logger.info("[{0}] Training model".format(self.name))
            self.logger.info(
                "[{0}] n_train={1}  #epochs={2}  batch size={3}".format(
                    self.name, n, self.settings["n_epochs"], batch_size
                )
            )

        dev_score_opt = 0.0
        dev_score_epo = -1
        for epoch in range(self.settings["n_epochs"]):
            iteration_losses = []

            # Shuffle the training data
            idxs = self.rand_state.permutation(n)

            nn.Module.train(self, True)
            for batch_st in range(0, n, batch_size):
                # zero gradients for each batch
                self.optimizer.zero_grad()

                batch_ed = batch_st + batch_size if batch_st + batch_size <= n else n

                output = self._calc_logits(
                    [_X_train[idx] for idx in idxs[batch_st:batch_ed]], batch_size
                )

                # Calculate loss for current batch
                y = torch.Tensor(_Y_train[idxs[batch_st:batch_ed]])
                if self.settings["host_device"] in self._gpu:
                    y = y.cuda()
                loss = self.loss(output, y)

                # Compute gradient
                loss.backward()

                # Update the parameters
                self.optimizer.step()

                if self.settings["host_device"] in self._gpu:
                    iteration_losses.append(loss.cpu())
                else:
                    iteration_losses.append(loss)

            # Print training stats and optionally checkpoint model
            if verbose and (
                (
                    (epoch + 1) % print_freq == 0
                    or epoch in [0, (self.settings["n_epochs"] - 1)]
                )
            ):
                msg = "[{0}] Epoch {1} ({2:.2f}s)\tAverage loss={3:.6f}".format(
                    self.name,
                    epoch + 1,
                    time() - st,
                    torch.stack(iteration_losses).mean(),
                )
                if X_dev is not None:
                    scores = self.score(
                        _X_dev, Y_dev, batch_size=self.settings["batch_size"]
                    )
                    score = scores if self.cardinality > 2 else scores[-1]
                    score_label = "Acc." if self.cardinality > 2 else "F1"
                    msg += "\tDev {0}={1:.2f}".format(score_label, 100.0 * score)
                self.logger.info(msg)

                # If best score on dev set so far and dev checkpointing is
                # active, save checkpoint
                if (
                    X_dev is not None
                    and dev_ckpt
                    and epoch > dev_ckpt_delay * self.settings["n_epochs"]
                    and score > dev_score_opt
                ):
                    dev_score_opt = score
                    dev_score_epo = epoch
                    self.save(save_dir=save_dir, global_step=dev_score_epo)

        # Conclude training
        if verbose:
            self.logger.info(
                "[{0}] Training done ({1:.2f}s)".format(self.name, time() - st)
            )
        # If checkpointing on, load last checkpoint (i.e. best on dev set)
        if dev_ckpt and X_dev is not None and verbose and dev_score_opt > 0:
            self.logger.info("Loading best checkpoint")
            self.load(save_dir=save_dir, global_step=dev_score_epo)

    def marginals(self, X, batch_size=None):
        """
        Compute the marginals for the given candidates X.
        Note: split into batches to avoid OOM errors.

        :param X: The input data which is a (list of Candidate objects, a sparse
            matrix of corresponding features) pair or a list of
            (Candidate, features) pairs.
        :type X: pair or list
        :param batch_size: Batch size.
        :type batch_size: int
        """
        nn.Module.train(self, False)

        if self._check_input(X):
            X = self._preprocess_data(X)

        marginal = self._calc_logits(X, batch_size)

        if self.settings["host_device"] in self._gpu:
            marginal = marginal.cpu()

        if self.cardinality == 2:
            return torch.sigmoid(marginal).detach().numpy()
        else:
            return torch.softmax(marginal).detach().numpy()

    def save(
        self, model_name=None, save_dir="checkpoints", verbose=True, global_step=0
    ):
        """Save current model.

        :param model_name: Saved model name.
        :type model_name: str
        :param save_dir: Saved model directory.
        :type save_dir: str
        :param verbose: Print log or not
        :type verbose: bool
        :param global_step: learned epoch of saved model
        :type global_step: int
        """

        model_name = model_name or self.name

        # Check existence of model saving directory and create if does not exist.
        model_dir = os.path.join(save_dir, model_name)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        params = {
            "model": self.state_dict(),
            "cardinality": self.cardinality,
            "name": model_name,
            "config": self.settings,
            "epoch": global_step,
        }

        model_file = "{0}.mdl.ckpt.{1}".format(model_name, global_step)

        try:
            torch.save(params, "{0}/{1}".format(model_dir, model_file))
        except BaseException:
            self.logger.warning("Saving failed... continuing anyway.")

        if verbose:
            self.logger.info(
                "[{0}] Model saved as {1} in {2}".format(
                    model_name, model_file, model_dir
                )
            )

    def load(
        self, model_name=None, save_dir="checkpoints", verbose=True, global_step=0
    ):
        """Load model from file and rebuild the model.

        :param model_name: Saved model name.
        :type model_name: str
        :param save_dir: Saved model directory.
        :type save_dir: str
        :param verbose: Print log or not
        :type verbose: bool
        :param global_step: learned epoch of saved model
        :type global_step: int
        """

        model_name = model_name or self.name

        model_dir = os.path.join(save_dir, model_name)
        if not os.path.exists(model_dir):
            self.logger.error("Loading failed... Directory does not exist.")

        model_file = "{0}.mdl.ckpt.{1}".format(model_name, global_step)

        try:
            checkpoint = torch.load("{0}/{1}".format(model_dir, model_file))
        except BaseException:
            self.logger.error(
                "Loading failed... Cannot load model from {0} in {1}".format(
                    model_name, model_dir
                )
            )

        self.load_state_dict(checkpoint["model"])
        self.settings = checkpoint["config"]
        self.cardinality = checkpoint["cardinality"]
        self.name = checkpoint["name"]

        if verbose:
            self.logger.info(
                "[{0}] Model loaded as {1} in {2}".format(
                    model_name, model_file, model_dir
                )
            )



import logging

from scipy.sparse import csr_matrix
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm.exc import NoResultFound

from fonduer.candidates.models import Candidate

logger = logging.getLogger(__name__)

# Flag value to signal that no filtering on split should be applied. Not an
# integer to ensure that it won't conflict with a user's split value.
ALL_SPLITS = "ALL"


def _get_cand_values(candidate, key_table):
    """Get the corresponding values for the key_table."""
    # NOTE: Import just before checking to avoid circular imports.
    from fonduer.features.models import FeatureKey
    from fonduer.supervision.models import GoldLabelKey, LabelKey

    if key_table == FeatureKey:
        return candidate.features
    elif key_table == LabelKey:
        return candidate.labels
    elif key_table == GoldLabelKey:
        return candidate.gold_labels
    else:
        raise ValueError("{} is not a valid key table.".format(key_table))


def _batch_postgres_query(table, records):
    """Break the list into chunks that can be processed as a single statement.

    Postgres query cannot be too long or it will fail.
    See: https://dba.stackexchange.com/questions/131399/is-there-a-maximum-
     length-constraint-for-a-postgres-query

    :param records: The full list of records to batch.
    :type records: iterable
    :param table: The sqlalchemy table.
    :return: A generator of lists of records.
    """
    if not records:
        return

    POSTGRESQL_MAX = 0x3FFFFFFF

    # Create preamble and measure its length
    preamble = (
        "INSERT INTO "
        + table.__tablename__
        + " ("
        + ", ".join(records[0].keys())
        + ") VALUES ("
        + ", ".join(["?"] * len(records[0].keys()))
        + ")\n"
    )
    start = 0
    end = 0
    total_len = len(preamble)
    while end < len(records):
        record_len = sum([len(str(v)) for v in records[end].values()])

        # Pre-increment to include the end element in the slice
        end += 1

        if total_len + record_len >= POSTGRESQL_MAX:
            logger.debug("Splitting query due to length ({} chars).".format(total_len))
            yield records[start:end]
            start = end
            # Reset the total query length
            total_len = len(preamble)
        else:
            total_len += record_len

    yield records[start:end]


def get_sparse_matrix_keys(session, key_table):
    """Return a generator of keys for the sparse matrix."""
    return session.query(key_table).order_by(key_table.name).all()


def batch_upsert_records(session, table, records):
    """Batch upsert records into postgresql database."""
    if not records:
        return
    for record_batch in _batch_postgres_query(table, records):
        stmt = insert(table.__table__)
        stmt = stmt.on_conflict_do_update(
            constraint=table.__table__.primary_key,
            set_={
                "keys": stmt.excluded.get("keys"),
                "values": stmt.excluded.get("values"),
            },
        )
        session.execute(stmt, record_batch)
        session.commit()


def get_sparse_matrix(session, key_table, cand_lists, key=None):
    """Load sparse matrix of GoldLabels for each candidate_class."""
    result = []
    cand_lists = cand_lists if isinstance(cand_lists, (list, tuple)) else [cand_lists]

    # Keys are used as a global index
    if key:
        keys_map = {key: 0}
    else:
        keys_map = {
            key.name: idx
            for (idx, key) in enumerate(get_sparse_matrix_keys(session, key_table))
        }

    for cand_list in cand_lists:
        indptr = [0]
        indices = []
        data = []
        for cand in cand_list:
            values = _get_cand_values(cand, key_table)
            if values:
                for cand_key, cand_value in zip(values[0].keys, values[0].values):
                    if cand_key in keys_map:
                        indices.append(keys_map[cand_key])
                        data.append(cand_value)

            indptr.append(len(indices))

        result.append(
            csr_matrix((data, indices, indptr), shape=(len(cand_list), len(keys_map)))
        )

    return result


def get_docs_from_split(session, candidate_classes, split):
    """Return a list of documents that contain the candidates in the split."""
    # Only grab the docs containing candidates from the given split.
    sub_query = session.query(Candidate.id).filter(Candidate.split == split).subquery()
    split_docs = set()
    for candidate_class in candidate_classes:
        split_docs.update(
            cand.document
            for cand in session.query(candidate_class)
            .filter(candidate_class.id.in_(sub_query))
            .all()
        )
    return split_docs


def get_mapping(session, table, candidates, generator, key_set):
    """Generate map of keys and values for the candidate from the generator.

    :param session: The database session.
    :param table: The table we will be inserting into (i.e. Feature or Label).
    :param candidates: The candidates to get mappings for.
    :param generator: A generator yielding (candidate_id, key, value) tuples.
    :param key_set: A mutable set which keys will be added to.
    :return: Dictionary of {"candidate_id": _, "keys": _, "values": _}
    :rtype: dict
    """
    for cand in candidates:
        # Grab the old values currently in the DB
        try:
            temp = session.query(table).filter(table.candidate_id == cand.id).one()
            cand_map = dict(zip(temp.keys, temp.values))
        except NoResultFound:
            cand_map = {}

        map_args = {"candidate_id": cand.id}
        for cid, key, value in generator(cand):
            if value == 0:
                continue
            cand_map[key] = value

        # Assemble label arguments
        map_args["keys"] = [*cand_map.keys()]
        map_args["values"] = [*cand_map.values()]

        # mutate the passed in key_set
        key_set.update(map_args["keys"])
        yield map_args


def get_cands_list_from_split(session, candidate_classes, doc, split):
    """Return the list of list of candidates from this document based on the split."""
    cands = []
    if split == ALL_SPLITS:
        # Get cands from all splits
        for candidate_class in candidate_classes:
            cands.append(
                session.query(candidate_class)
                .filter(candidate_class.document_id == doc.id)
                .all()
            )
    else:
        # Get cands from the specified split
        for candidate_class in candidate_classes:
            cands.append(
                session.query(candidate_class)
                .filter(candidate_class.document_id == doc.id)
                .filter(candidate_class.split == split)
                .all()
            )
    return cands


def add_keys(session, key_table, keys):
    """Bulk add annotation keys to the specified table.

    :param table: The sqlalchemy class to insert into.
    :param keys: A list of strings to insert into the table.
    """
    # Do nothing if empty
    if not keys:
        return

    for key_batch in _batch_postgres_query(key_table, [{"name": key} for key in keys]):
        # Rather than deal with concurrency of querying first then inserting only
        # new keys, insert with on_conflict_do_nothing.
        stmt = insert(key_table.__table__)
        stmt = stmt.on_conflict_do_nothing(constraint=key_table.__table__.primary_key)
        session.execute(stmt, key_batch)
